# encoding: utf-8
"""
sradj.py

Created by Evelio Vila
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

import json
from struct import unpack
from exabgp.vendoring import six

from exabgp.vendoring.bitstring import BitArray
from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE, LsGenericFlags

#    draft-gredler-idr-bgp-ls-segment-routing-ext-03
#    0                   1                   2                   3
#    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |               Type            |              Length           |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   | Flags         |     Weight    |             Reserved          |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |                   SID/Label/Index (variable)                  |
#   +---------------------------------------------------------------+
#

@LINKSTATE.register()
class SrAdjacency(object):
	TLV = 1099

	def __init__ (self, flags, sids, weight):
		self.flags = flags
		self.sids = sids
		self.weight = weight

	def __repr__ (self):
		return "adj_flags: %s, sids: %s" % (self.flags, self.sids)

	@classmethod
	def unpack (cls,data,length):
		# We only support IS-IS flags for now.
		flags = LsGenericFlags.unpack(data[0:1],LsGenericFlags.ISIS_SR_ADJ_FLAGS)
		# Parse adj weight
		weight = six.indexbytes(data,1)
		# Move pointer 4 bytes: Flags(1) + Weight(1) + Reserved(2)
		data = data[4:]
     	# SID/Index/Label: according to the V and L flags, it contains
      	# either:
		# *  A 3 octet local label where the 20 rightmost bits are used for
		#	 encoding the label value.  In this case the V and L flags MUST
		#	 be set.
		#
		# *  A 4 octet index defining the offset in the SID/Label space
		# 	 advertised by this router using the encodings defined in
		#  	 Section 3.1.  In this case V and L flags MUST be unset.
		sids = []
		while data:
			# Range Size: 3 octet value indicating the number of labels in
			# the range.
			if int(flags.flags['V']) and int(flags.flags['L']):
				b = BitArray(bytes=data[:3])
				sid = b.unpack('uintbe:24')[0:1]
				data = data[3:]
			elif (not flags.flags['V']) and \
				(not flags.flags['L']):
				sid = unpack('!I',data[:4])[0:1]
				data = data[4:]
			sids.append(sid)
		return cls(flags=flags, sids=sids, weight=weight)

	def json (self,compact=None):
		return ', '.join(['"sr-adj-flags": {}'.format(self.flags.json()),
			'"sids": {}'.format(json.dumps(self.sids)),
			'"sr-adj-weight": {}'.format(json.dumps(self.weight))])



# encoding: utf-8
"""
announce/__init__.py

Created by Thomas Mangin on 2015-06-04.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
"""

from exabgp.util import ordinal
from exabgp.util import character
from exabgp.util import concat_bytes_i

from exabgp.protocol.ip import IP

from exabgp.bgp.message import OUT
from exabgp.rib.change import Change

from exabgp.protocol.family import AFI
from exabgp.configuration.core import Section

from exabgp.bgp.message.update.nlri.cidr import CIDR
from exabgp.bgp.message.update.attribute import Attribute


# Take an integer an created it networked packed representation for the right family (ipv4/ipv6)
def pack_int (afi, integer):
	return concat_bytes_i(character((integer >> (offset * 8)) & 0xff) for offset in range(IP.length(afi)-1,-1,-1))


class ParseAnnounce (Section):
	def post (self):
		self._split()
		routes = self.scope.pop(self.name)
		if routes:
			self.scope.extend_routes(routes)
		return True

	@staticmethod
	def split (last):
		if Attribute.CODE.INTERNAL_SPLIT not in last.attributes:
			yield last
			return

		# ignore if the request is for an aggregate, or the same size
		mask = last.nlri.cidr.mask
		cut = last.attributes[Attribute.CODE.INTERNAL_SPLIT]
		if mask >= cut:
			yield last
			return

		# calculate the number of IP in the /<size> of the new route
		increment = pow(2,last.nlri.afi.mask() - cut)
		# how many new routes are we going to create from the initial one
		number = pow(2,cut - last.nlri.cidr.mask)

		# convert the IP into a integer/long
		ip = 0
		for c in last.nlri.cidr.ton():
			ip <<= 8
			ip += ordinal(c)

		afi = last.nlri.afi
		safi = last.nlri.safi

		# Really ugly
		klass = last.nlri.__class__
		path_info = last.nlri.path_info
		nexthop = last.nlri.nexthop

		# XXX: Looks weird to set and then set to None, check
		last.nlri.cidr.mask = cut
		last.nlri = None

		# generate the new routes
		for _ in range(number):
			# update ip to the next route, this recalculate the "ip" field of the Inet class
			nlri = klass(afi,safi,OUT.ANNOUNCE)
			nlri.cidr = CIDR(pack_int(afi,ip),cut)
			nlri.nexthop = nexthop  # nexthop can be NextHopSelf
			nlri.path_info = path_info
			# next ip
			ip += increment
			yield Change(nlri,last.attributes)

	def _split (self):
		for route in self.scope.pop_routes():
			for splat in self.split(route):
				self.scope.append_route(splat)


class SectionAnnounce (ParseAnnounce):
	name = 'announce'

	def __init__ (self, tokeniser, scope, error, logger):
		Section.__init__(self,tokeniser,scope,error,logger)

	def clear (self):
		return True

	def pre (self):
		return True

	def post (self):
		routes = self.scope.pop('routes',[])
		self.scope.pop()
		if routes:
			self.scope.extend('routes',routes)
		self.scope.pop(self.name)
		return True


class AnnounceIPv4 (ParseAnnounce):
	name = 'ipv4'
	afi = AFI.ipv4

	def clear (self):
		return True


class AnnounceIPv6 (ParseAnnounce):
	name = 'ipv6'
	afi = AFI.ipv6

	def clear (self):
		return True


class AnnounceL2VPN (ParseAnnounce):
	name = 'l2vpn'
	afi = AFI.l2vpn

	def clear (self):
		return True



"""
nlri.py

Created by Evelio Vila on 2016-11-26. eveliovila@gmail.com
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI

from exabgp.util import ordinal
from exabgp.bgp.message import OUT

from exabgp.bgp.message.update.nlri import NLRI
from exabgp.bgp.message.update.nlri.qualifier import RouteDistinguisher

# https://tools.ietf.org/html/rfc7752#section-3.2
#
#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |            NLRI Type          |     Total NLRI Length         |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |                                                               |
#     //                  Link-State NLRI (variable)                 //
#     |                                                               |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#
#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |            NLRI Type          |     Total NLRI Length         |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |                                                               |
#     +                       Route Distinguisher                     +
#     |                                                               |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |                                                               |
#     //                  Link-State NLRI (variable)                 //
#     |                                                               |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#
#                   +------+---------------------------+
#                   | Type | NLRI Type                 |
#                   +------+---------------------------+
#                   |  1   | Node NLRI                 |
#                   |  2   | Link NLRI                 |
#                   |  3   | IPv4 Topology Prefix NLRI |
#                   |  4   | IPv6 Topology Prefix NLRI |
#                   +------+---------------------------+
# ==================================================================== BGP LINK_STATE
#            +-------------+----------------------------------+
#            | Protocol-ID | NLRI information source protocol |
#            +-------------+----------------------------------+
#            |      1      | IS-IS Level 1                    |
#            |      2      | IS-IS Level 2                    |
#            |      3      | OSPFv2                           |
#            |      4      | Direct                           |
#            |      5      | Static configuration             |
#            |      6      | OSPFv3                           |
#            +-------------+----------------------------------+
# ===================================================================== PROTO_ID

PROTO_CODES = {
	1:	'isis_l1',
	2:	'isis_l2',
	3:	'ospf_v2',
	4:	'direct',
	5:	'static',
	6:	'ospfv3',
}


@NLRI.register(AFI.bgpls,SAFI.bgp_ls)
@NLRI.register(AFI.bgpls,SAFI.bgp_ls_vpn)
class BGPLS(NLRI):
	registered_bgpls = dict()

	CODE = -1
	NAME = 'Unknown'
	SHORT_NAME = 'unknown'

	def __init__(self, action=OUT.UNSET, addpath=None):
		NLRI.__init__(self, AFI.bgpls, SAFI.bgp_ls, action)
		self._packed = b''

	def pack_nlri(self, negotiated=None):
		return pack('!BB',self.CODE,len(self._packed)) + self._packed

	def __len__(self):
		return len(self._packed) + 2

	def __hash__(self):
		return hash("%s:%s:%s:%s" % (self.afi,self.safi,self.CODE,self._packed))

	def __str__(self):
		return "bgp-ls:%s:%s" % (self.registered_bgpls.get(self.CODE,self).SHORT_NAME.lower(),'0x' + ''.join('%02x' % ordinal(_) for _ in self._packed))

	@classmethod
	def register(cls, klass):
		if klass.CODE in cls.registered_bgpls:
			raise RuntimeError('only one BGP LINK_STATE registration allowed')
		cls.registered_bgpls[klass.CODE] = klass
		return klass

	@classmethod
	def unpack_nlri(cls, afi, safi, bgp, action, addpath):
		code, length = unpack('!HH',bgp[:4])
		if code in cls.registered_bgpls:
			if safi == SAFI.bgp_ls_vpn:
				# Extract Route Distinguisher
				rd = RouteDistinguisher.unpack(bgp[4:12])
				klass = cls.registered_bgpls[code].unpack_nlri(bgp[12:length+4],rd)
			else:
				rd = None
				klass = cls.registered_bgpls[code].unpack_nlri(bgp[4:length+4],rd)
		else:
			klass = GenericBGPLS(code,bgp[4:length+4])
		klass.CODE = code
		klass.action = action
		klass.addpath = addpath

		return klass,bgp[length+4:]

	def _raw(self):
		return ''.join('%02X' % ordinal(_) for _ in self.pack())


class GenericBGPLS(BGPLS):
	def __init__(self,code,packed):
		BGPLS.__init__(self)
		self.CODE = code
		self._pack(packed)

	def _pack(self,packed=None):
		if self._packed:
			return self._packed

		if packed:
			self._packed = packed
			return packed

	def json(self, compact=None):
		return '{ "code": %d, "parsed": false, "raw": "%s" }' % (self.CODE,self._raw())



"""
node.py

Created by Evelio Vila on 2016-11-26. eveliovila@gmail.com
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import unpack

from exabgp.bgp.message.update.nlri.bgpls.nlri import BGPLS
from exabgp.bgp.message.update.nlri.bgpls.nlri import PROTO_CODES
from exabgp.bgp.message.update.nlri.bgpls.tlvs.node import NodeDescriptor

#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+
#     |  Protocol-ID  |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |                           Identifier                          |
#     |                            (64 bits)                          |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     //                Local Node Descriptors (variable)            //
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# ===================================================================== NODENLRI
#             +------------+----------------------------------+
#             | Identifier | Routing Universe                 |
#             +------------+----------------------------------+
#             |     0      | Default Layer 3 Routing topology |
#             +------------+----------------------------------+
#   The Protocol-ID field can contain one of the following values:
# ===================================================================== DOMAIN


@BGPLS.register
class NODE(BGPLS):
	CODE = 1
	NAME = "Node NLRI"
	SHORT_NAME = "Node"

	def __init__ (
			self, domain, proto_id, node_ids,
			packed=None,nexthop=None,action=None,
			route_d=None,addpath=None):
		BGPLS.__init__(self,action,addpath)
		self.domain = domain
		self.proto_id = proto_id
		self.node_ids = node_ids
		self.nexthop = nexthop
		self._pack = packed
		self.route_d = route_d

	def json (self,compact=None):
		nodes = ', '.join(d.json() for d in self.node_ids)
		content = ', '.join([
			'"ls-nlri-type": 1',
			'"l3-routing-topology": %d' % int(self.domain),
			'"protocol-id": %d' % int(self.proto_id),
			'"node-descriptors": { %s }' % nodes,
			'"nexthop": "%s"' % self.nexthop,
		])
		if self.route_d:
			content += ', %s' % self.route_d.json()
		return '{ %s }' % (content)

	@classmethod
	def unpack_nlri (cls, data, rd):
		proto_id = unpack('!B',data[0:1])[0]
		if proto_id not in PROTO_CODES.keys():
			raise Exception('Protocol-ID {} is not valid'.format(proto_id))
		domain = unpack('!Q',data[1:9])[0]

		# unpack list of node descriptors
		node_type, node_length = unpack('!HH',data[9:13])
		if node_type != 256:
			raise Exception(
				'Unknown type: {}. Only Local Node descriptors are allowed in'
				'Node type msg'.format(node_type)
			)
		values = data[13: 13 + node_length]

		node_ids = []
		while values:
			# Unpack Node Descriptor Sub-TLVs
			node_id, left = NodeDescriptor.unpack(values, proto_id)
			node_ids.append(node_id)
			if left == values:
				raise RuntimeError("sub-calls should consume data")
			values = left

		return cls(
			domain=domain,proto_id=proto_id,
			node_ids=node_ids,route_d=rd,packed=data
		)

	def __eq__ (self, other):
		return \
			isinstance(other, BGPLS) and \
			self.CODE == other.CODE and \
			self.domain == other.domain and \
			self.proto_id == other.proto_id and \
			self.route_d == other.route_d

	def __ne__ (self, other):
		return not self.__eq__(other)

	def __str__ (self):
		return self.json()

	def __hash__ (self):
		return hash((self.proto_id,self.node_ids))

	def pack (self,negotiated=None):
		return self._pack



# encoding: utf-8
"""
section.py

Created by Thomas Mangin on 2015-06-04.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from string import ascii_letters
from string import digits

from exabgp.configuration.core.error import Error


class Section (Error):
	name = 'undefined'
	known = dict()     # command/section and code to handle it
	default = dict()   # command/section has a a defult value, use it if no data was provided
	action = {}        # how to handle this command ( append, add, assign, route )
	assign = {}        # configuration to class variable lookup for setattr

	def __init__ (self, tokerniser, scope, error, logger):
		Error.__init__(self)
		self.tokeniser = tokerniser
		self.scope = scope
		self.error = error
		self.logger = logger
		self._names = []

	def clear (self):
		raise RuntimeError('%s did not implemented clear as should be' % self.__class__.__name__)

	@classmethod
	def register (cls, name, action,afi=''):
		def inner (function):
			identifier = (afi,name) if afi else name
			if identifier in cls.known:
				raise RuntimeError('more than one registration per command attempted')
			cls.known[identifier] = function
			cls.action[identifier] = action
			return function
		return inner

	def check_name (self, name):
		if any(False if c in ascii_letters + digits + '.-_' else True for c in name):
			self.throw('invalid character in name for %s ' % self.name)
		if name in self._names:
			self.throw('the name "%s" already exists in %s' % (name,self.name))
		self._names.append(name)

	def pre (self):
		return True

	def post (self):
		return True

	def parse (self, name, command):
		identifier = command if command in self.known else (self.name,command)
		if identifier not in self.known:
			return self.error.set('unknown command %s options are %s' % (command,', '.join([str(_) for _ in self.known])))

		try:
			if command in self.default:
				insert = self.known[identifier](self.tokeniser.iterate,self.default[command])
			else:
				insert = self.known[identifier](self.tokeniser.iterate)

			action = self.action.get(identifier,'')

			if action == 'set-command':
				self.scope.set(command,insert)
			elif action == 'extend-name':
				self.scope.extend(name,insert)
			elif action == 'append-name':
				self.scope.append(name,insert)
			elif action == 'append-command':
				self.scope.append(command,insert)
			elif action == 'extend-command':
				self.scope.extend(command,insert)
			elif action == 'attribute-add':
				self.scope.attribute_add(name,insert)
			elif action == 'nlri-set':
				self.scope.nlri_assign(name,self.assign[command],insert)
			elif action == 'nlri-add':
				for adding in insert:
					self.scope.nlri_add(name,command,adding)
			elif action == 'nlri-nexthop':
				self.scope.nlri_nexthop(name,insert)
			elif action == 'nexthop-and-attribute':
				ip, attribute = insert
				if ip:
					self.scope.nlri_nexthop(name,ip)
				if attribute:
					self.scope.attribute_add(name,attribute)
			elif action == 'append-route':
				self.scope.extend_routes(insert)
			elif action == 'nop':
				pass
			else:
				raise RuntimeError('name %s command %s has no action set' % (name,command))
			return True
		except ValueError as exc:
			return self.error.set(str(exc))

		return True



# encoding: utf-8
"""
linkid.py

Created by Evelio Vila on 2016-11-26. eveliovila@gmail.com
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack

from exabgp.bgp.message.notification import Notify
from exabgp.util import ordinal

#       0                   1                   2                   3
#       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      |                  Link Local Identifier                        |
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      |                  Link Remote Identifier                       |
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      https://tools.ietf.org/html/rfc5307 sec 1.1
# ================================================================== Link Local/Remote Identifiers


class LinkIdentifier (object):
	def __init__ (self, local_id, remote_id, packed=None):
		self.local_id = local_id
		self.remote_id = remote_id
		self._packed = packed

	@classmethod
	def unpack (cls, data):
		local_id = unpack('!L',data[:4])[0]
		remote_id = unpack('!L',data[4:8])[0]
		return cls(local_id=local_id,remote_id=remote_id)

	def json (self):
		content = '"link-local-id": %s, ' % self.local_id + \
			'"link-remote-id": %s' % self.remote_id
		return content

	def __eq__ (self, other):
		return (self.local_id == other.local_id) and (self.remote_id == other.remote_id)

	def __neq__ (self, other):
		return self.local_id != other.local_id

	def __lt__ (self, other):
		raise RuntimeError('Not implemented')

	def __le__ (self, other):
		raise RuntimeError('Not implemented')

	def __gt__ (self, other):
		raise RuntimeError('Not implemented')

	def __ge__ (self, other):
		raise RuntimeError('Not implemented')

	def __str__ (self):
		return ':'.join('%02X' % ordinal(_) for _ in self._packed)

	def __repr__ (self):
		return self.__str__()

	def __len__ (self):
		return len(self._packed)

	def __hash__ (self):
		return hash(str(self))

	def pack (self):
		return self._packed



# encoding: utf-8
"""
peer.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2017-2017 Exa Networks. All rights reserved.
"""

from exabgp.vendoring import six

from exabgp.bgp.timer import SendTimer
from exabgp.bgp.message import Notify

from exabgp.reactor.network.error import NetworkError


# =========================================================================== KA
#

class KA (object):
	def __init__ (self, session, proto):
		self._generator = self._keepalive(proto)
		self.send_timer = SendTimer(session,proto.negotiated.holdtime)

	def _keepalive (self, proto):
		need_ka   = False
		generator = None

		while True:
			# SEND KEEPALIVES
			need_ka |= self.send_timer.need_ka()

			if need_ka:
				if not generator:
					generator = proto.new_keepalive()
					need_ka = False

			if not generator:
				yield False
				continue

			try:
				# try to close the generator and raise a StopIteration in one call
				six.next(generator)
				six.next(generator)
				# still running
				yield True
			except NetworkError:
				raise Notify(4,0,'problem with network while trying to send keepalive')
			except StopIteration:
				generator = None
				yield False

	def __call__ (self):
		#  True  if we need or are trying
		#  False if we do not need to send one
		try:
			return six.next(self._generator)
		except StopIteration:
			raise Notify(4,0,'could not send keepalive')



# encoding: utf-8
"""
__init__.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import unpack

from exabgp.util import character
from exabgp.util import ordinal
from exabgp.util import concat_bytes
from exabgp.util import bytes_ascii
from exabgp.bgp.message.message import Message
from exabgp.bgp.message.notification import Notify

from exabgp.bgp.message.open.version import Version
from exabgp.bgp.message.open.asn import ASN
from exabgp.bgp.message.open.holdtime import HoldTime
from exabgp.bgp.message.open.routerid import RouterID
from exabgp.bgp.message.open.capability import Capabilities

# =================================================================== Open

# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+
# |    Version    |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |     My Autonomous System      |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |           Hold Time           |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                         BGP Identifier                        |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# | Opt Parm Len  |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                                                               |
# |             Optional Parameters (variable)                    |
# |                                                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

# Optional Parameters:

# 0                   1
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-...
# |  Parm. Type   | Parm. Length  |  Parameter Value (variable)
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-...


@Message.register
class Open (Message):
	ID = Message.CODE.OPEN
	TYPE = character(Message.CODE.OPEN)

	def __init__ (self, version, asn, hold_time, router_id, capabilities):
		self.version = version
		self.asn = asn
		self.hold_time = hold_time
		self.router_id = router_id
		self.capabilities = capabilities

	def message (self,negotiated=None):
		return self._message(concat_bytes(
			self.version.pack(),
			self.asn.trans().pack(),
			self.hold_time.pack(),
			self.router_id.pack(),
			self.capabilities.pack()
		))

	def __str__ (self):
		return "OPEN version=%d asn=%d hold_time=%s router_id=%s capabilities=[%s]" % (self.version, self.asn.trans(), self.hold_time, self.router_id,self.capabilities)

	@classmethod
	def unpack_message (cls, data, _=None):
		version = ordinal(data[0])
		if version != 4:
			# Only version 4 is supported nowdays..
			raise Notify(2,1,bytes_ascii(data[0]))
		asn = unpack('!H',data[1:3])[0]
		hold_time = unpack('!H',data[3:5])[0]
		numeric = unpack('!L',data[5:9])[0]
		router_id = "%d.%d.%d.%d" % (numeric >> 24,(numeric >> 16) & 0xFF,(numeric >> 8) & 0xFF,numeric & 0xFF)
		return cls(
			Version(version),
			ASN(asn),
			HoldTime(hold_time),
			RouterID(router_id),
			Capabilities.unpack(data[9:])
		)



# encoding: utf-8
"""
srprefix.py

Created by Evelio Vila
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

import json
from struct import unpack
from exabgp.vendoring import six

from exabgp.vendoring.bitstring import BitArray
from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE, LsGenericFlags

#    draft-gredler-idr-bgp-ls-segment-routing-ext-03
#    0                   1                   2                   3
#    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |               Type            |            Length             |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |     Flags       |  Algorithm  |           Reserved            |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |                       SID/Index/Label (variable)              |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

@LINKSTATE.register()
class SrPrefix(object):
	TLV = 1158

	def __init__ (self, flags, sids, sr_algo):
		self.flags = flags
		self.sids = sids
		self.sr_algo = sr_algo

	def __repr__ (self):
		return "prefix_flags: %s, sids: %s" % (self.flags, self.sids)

	@classmethod
	def unpack (cls,data,length):
		# We only support IS-IS flags for now.
		flags = LsGenericFlags.unpack(data[0:1],LsGenericFlags.ISIS_SR_FLAGS)
		#
		# Parse Algorithm
		sr_algo = six.indexbytes(data, 1)
		# Move pointer 4 bytes: Flags(1) + Algorithm(1) + Reserved(2)
		data = data[4:]
     	# SID/Index/Label: according to the V and L flags, it contains
      	# either:
		# *  A 3 octet local label where the 20 rightmost bits are used for
		#	 encoding the label value.  In this case the V and L flags MUST
		#	 be set.
		#
		# *  A 4 octet index defining the offset in the SID/Label space
		# 	 advertised by this router using the encodings defined in
		#  	 Section 3.1.  In this case V and L flags MUST be unset.
		sids = []
		while data:
			if flags.flags['V'] and flags.flags['L']:
				b = BitArray(bytes=data[:3])
				sid = b.unpack('uintbe:24')[0]
				data = data[3:]
			elif (not flags.flags['V']) and \
				(not flags.flags['L']):
				if len(data) != 4:
    				# Cisco IOS XR Software, Version 6.1.1.19I is not
					# correctly setting the flags
					raise Notify(3,5, "SID/Label size doesn't match V and L flag state")
				sid = unpack('!I', data[:4])[0]
				data = data[4:]
			sids.append(sid)
		return cls(flags=flags, sids=sids, sr_algo=sr_algo)

	def json (self,compact=None):
		return ', '.join(['"sr-prefix-flags": {}'.format(self.flags.json()),
			'"sids": {}'.format(json.dumps(self.sids)),
			'"sr-algorithm": {}'.format(json.dumps(self.sr_algo))])



#!/usr/bin/env python
# encoding: utf-8
"""
open.py

Created by Thomas Morin, Orange on 2015-07-10.
Copyright (c) 2009-2015 Orange. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import unittest

from exabgp.util import character

from exabgp.reactor.protocol import AFI, SAFI

from exabgp.bgp.message.update import Attributes

from exabgp.bgp.message.update.attribute.localpref import LocalPreference
from exabgp.bgp.message.update.attribute.community.extended.communities \
    import ExtendedCommunities
from exabgp.bgp.message.update.attribute.community.extended \
    import RouteTargetASN2Number as RouteTarget
from exabgp.bgp.message.update.attribute.community.extended.encapsulation \
    import Encapsulation
from exabgp.bgp.message.update.attribute.community.extended \
    import RTRecord

from exabgp.bgp.message.update.nlri.ipvpn import IPVPN

from exabgp.bgp.message.update.nlri.rtc import RTC

from exabgp.bgp.message.update.nlri.evpn.mac import MAC as EVPNMAC
from exabgp.bgp.message.update.nlri.evpn.multicast import Multicast as EVPNMulticast
from exabgp.bgp.message.update.nlri.evpn.prefix import Prefix as EVPNPrefix
from exabgp.bgp.message.update.nlri.evpn.nlri import EVPN

from exabgp.bgp.message.update.nlri.qualifier.rd import RouteDistinguisher
from exabgp.bgp.message.update.nlri.qualifier.labels import Labels
from exabgp.bgp.message.update.nlri.qualifier.esi import ESI
from exabgp.bgp.message.update.nlri.qualifier.etag import EthernetTag
from exabgp.bgp.message.update.nlri.qualifier.mac import MAC

from exabgp.protocol.ip import IP

from exabgp.bgp.message import OUT

from exabgp.configuration.setup import environment
environment.setup('')


class TestNLRIs(unittest.TestCase):

    # Tests on IPVPN NLRIs

    def test200_IPVPNCreatePackUnpack(self):
        '''Test pack/unpack for IPVPN routes'''
        nlri = IPVPN.new(AFI.ipv4, SAFI.mpls_vpn,
                         IP.pton("1.2.3.0"), 24,
                         Labels([42], True),
                         RouteDistinguisher.fromElements("42.42.42.42", 5))

        packed = nlri.pack()
        unpacked,leftover = IPVPN.unpack_nlri(AFI.ipv4, SAFI.mpls_vpn,
                                              packed, OUT.UNSET, None)

        self.assertEqual(0, len(leftover))

        # TODO: compare packed with a reference encoding verified
        # as conformant with RFC4364

        self.assertTrue(isinstance(unpacked, IPVPN))

        self.assertEqual("1.2.3.0/24", unpacked.cidr.prefix())
        self.assertEqual(1, len(unpacked.labels.labels))
        self.assertEqual(42, unpacked.labels.labels[0])
        self.assertEqual("42.42.42.42:5", unpacked.rd._str())

    # Tests on EVPN NLRIs

    def test99_EVPNMACCreatePackUnpack(self):
        '''Test pack/unpack for E-VPN MAC routes'''
        nlri = EVPNMAC(RouteDistinguisher.fromElements("42.42.42.42", 5),
                       ESI(),
                       EthernetTag(111),
                       MAC("01:02:03:04:05:06"), 6*8,
                       Labels([42], True),
                       IP.create("1.1.1.1"))

        packed = nlri.pack()

        unpacked,leftover = EVPN.unpack_nlri(AFI.l2vpn, SAFI.evpn,
                                             packed, OUT.UNSET, None)

        self.assertEqual(0, len(leftover))

        # TODO: compare packed with a reference encoding verified
        # as conformant with RFC7432

        self.assertTrue(isinstance(unpacked, EVPNMAC))

        self.assertEqual("42.42.42.42:5", unpacked.rd._str())
        self.assertEqual(ESI.DEFAULT, unpacked.esi.esi)
        self.assertEqual(EthernetTag(111), unpacked.etag)
        self.assertEqual(MAC("01:02:03:04:05:06"), unpacked.mac)

        self.assertEqual(IP.create("1.1.1.1"), unpacked.ip)

        self.assertEqual(1, len(unpacked.label.labels))
        self.assertEqual(42, unpacked.label.labels[0])

    def test99_EVPNMulticastCreatePackUnpack(self):
        '''Test pack/unpack for E-VPN Multicast routes'''

        nlri = EVPNMulticast(RouteDistinguisher.fromElements("42.42.42.42", 5),
                             EthernetTag(111),
                             IP.create("1.1.1.1"))

        packed = nlri.pack()

        unpacked,leftover = EVPN.unpack_nlri(AFI.l2vpn, SAFI.evpn,
                                             packed, OUT.UNSET, None)

        self.assertEqual(0, len(leftover))

        # TODO: compare packed with a reference encoding verified
        # as conformant with RFC7432

        self.assertTrue(isinstance(unpacked, EVPNMulticast))

        self.assertEqual("42.42.42.42:5", unpacked.rd._str())
        self.assertEqual(EthernetTag(111), unpacked.etag)
        self.assertEqual(IP.create("1.1.1.1"), unpacked.ip)

    def test99_EVPNPrefixCreatePackUnpack(self):
        '''Test pack/unpack for E-VPN Prefix routes'''

        nlri = EVPNPrefix(RouteDistinguisher.fromElements("42.42.42.42", 5),
                          ESI(),
                          EthernetTag(111),
                          Labels([42], True),
                          IP.create("1.1.1.0"),24,
                          IP.create("2.2.2.2"),
                          )

        packed = nlri.pack()

        unpacked,leftover = EVPN.unpack_nlri(AFI.l2vpn, SAFI.evpn,
                                             packed, OUT.UNSET, None)

        self.assertEqual(0, len(leftover))

        # TODO: compare packed with a reference encoding verified
        # as conformant with RFC7432

        self.assertTrue(isinstance(unpacked, EVPNPrefix))

        self.assertEqual("42.42.42.42:5", unpacked.rd._str())
        self.assertEqual(ESI.DEFAULT, unpacked.esi.esi)
        self.assertEqual(EthernetTag(111), unpacked.etag)
        self.assertEqual(IP.create("1.1.1.0"), unpacked.ip)
        self.assertEqual(24, unpacked.iplen)
        self.assertEqual(IP.create("2.2.2.2"), unpacked.gwip)
        self.assertEqual(1, len(unpacked.label.labels))
        self.assertEqual(42, unpacked.label.labels[0])

    def test100_EVPNMACHashEqual(self):
        '''
        Two indistinct EVPN NLRI should
        hash to the same value, and be equal
        '''

        nlri1 = EVPNMAC(RouteDistinguisher.fromElements("42.42.42.42", 5),
                        ESI(),
                        EthernetTag(111),
                        MAC("01:02:03:04:05:06"), 6*8,
                        Labels([42], True),
                        IP.create("1.1.1.1"))

        nlri2 = EVPNMAC(RouteDistinguisher.fromElements("42.42.42.42", 5),
                        ESI(),
                        EthernetTag(111),
                        MAC("01:02:03:04:05:06"), 6*8,
                        Labels([42], True),
                        IP.create("1.1.1.1"))

        self.assertEqual(hash(nlri1), hash(nlri2))
        self.assertEqual(nlri1, nlri2)

    def test101_EVPNHashEqual_somefieldsvary(self):
        '''
        Two EVPN MAC NLRIs differing by their ESI or label or RD,
        or nexthop, but otherwise identical should hash to the same value,
        and be equal
        '''

        nlri0 = EVPNMAC(RouteDistinguisher.fromElements("42.42.42.42", 5),
                        ESI(),
                        EthernetTag(111),
                        MAC("01:02:03:04:05:06"), 6*8,
                        Labels([42], True),
                        IP.create("1.1.1.1"))

        # ESI
        nlri1 = EVPNMAC(RouteDistinguisher.fromElements("42.42.42.42", 5),
                        ESI(b''.join(character(1) for _ in range(0,10))),
                        EthernetTag(111),
                        MAC("01:02:03:04:05:06"), 6*8,
                        Labels([42], True),
                        IP.create("1.1.1.1"))

        # label
        nlri2 = EVPNMAC(RouteDistinguisher.fromElements("42.42.42.42", 5),
                        ESI(),
                        EthernetTag(111),
                        MAC("01:02:03:04:05:06"), 6*8,
                        Labels([4444], True),
                        IP.create("1.1.1.1"))

        # IP: different IPs, but same MACs: different route
        nlri3 = EVPNMAC(RouteDistinguisher.fromElements("42.42.42.42", 5),
                        ESI(),
                        EthernetTag(111),
                        MAC("01:02:03:04:05:06"), 6*8,
                        Labels([42], True),
                        IP.create("2.2.2.2"))

        # with a next hop...
        nlri4 = EVPNMAC(RouteDistinguisher.fromElements("42.42.42.42", 5),
                        ESI(),
                        EthernetTag(111),
                        MAC("01:02:03:04:05:06"), 6*8,
                        Labels([42], True),
                        IP.create("1.1.1.1"),
                        IP.pton("10.10.10.10"))
        nlri5 = EVPNMAC(RouteDistinguisher.fromElements("42.42.42.42", 5),
                        ESI(),
                        EthernetTag(111),
                        MAC("01:02:03:04:05:06"), 6*8,
                        Labels([42], True),
                        IP.create("1.1.1.1"),
                        IP.pton("11.11.11.11"))

        self.assertEqual(hash(nlri0), hash(nlri1))
        self.assertEqual(hash(nlri0), hash(nlri2))
        self.assertEqual(hash(nlri0), hash(nlri4))
        self.assertEqual(nlri0, nlri1)
        self.assertEqual(nlri0, nlri2)
        self.assertEqual(nlri0, nlri4)
        self.assertEqual(nlri1, nlri2)
        self.assertEqual(nlri1, nlri4)
        self.assertEqual(nlri2, nlri4)
        self.assertEqual(nlri4, nlri5)

        self.assertNotEqual(hash(nlri0), hash(nlri3))
        self.assertNotEqual(nlri0, nlri3)
        self.assertNotEqual(nlri1, nlri3)
        self.assertNotEqual(nlri2, nlri3)
        self.assertNotEqual(nlri3, nlri4)

    # tests on RTC NLRI

    def test99_RTCCreatePackUnpack(self):
        '''Test pack/unpack for RTC routes'''

        nlri = RTC.new(AFI.ipv4, SAFI.rtc,
                       64512,
                       RouteTarget(64577,123))

        packed = nlri.pack()
        unpacked,leftover = RTC.unpack_nlri(AFI.ipv4, SAFI.mpls_vpn,
                                            packed, OUT.UNSET, None)

        self.assertEqual(0, len(leftover))

        # TODO: compare packed with a reference encoding verified
        # as conformant with RFC4684

        self.assertTrue(isinstance(unpacked, RTC))

        self.assertEqual(64512, unpacked.origin)

        self.assertTrue(isinstance(unpacked.rt, RouteTarget))
        self.assertEqual(64577, unpacked.rt.asn)
        self.assertEqual(123, unpacked.rt.number)

    def test98_RTCWildcardPackUnpack(self):
        '''Test pack/unpack for RTC routes'''

        nlri = RTC.new(AFI.ipv4, SAFI.rtc,
                       0, None)

        packed = nlri.pack()
        unpacked,leftover = RTC.unpack_nlri(AFI.ipv4, SAFI.mpls_vpn,
                                            packed, OUT.UNSET, None)

        self.assertEqual(0, len(leftover))

        # TODO: compare packed with a reference encoding verified
        # as conformant with RFC4684

        self.assertTrue(isinstance(unpacked, RTC))

        self.assertEqual(0, unpacked.origin)

        self.assertIsNone(unpacked.rt)


    # tests on attributes

    def test4_DistinctAttributes(self):
        atts1 = Attributes()
        atts1.add(LocalPreference(10))

        atts2 = Attributes()
        atts2.add(LocalPreference(20))

        self.assertNotEqual(atts1, atts2)

    def test5_SameAttributes(self):
        atts1 = Attributes()
        atts1.add(LocalPreference(10))

        atts2 = Attributes()
        atts2.add(LocalPreference(10))

        self.assertEqual(hash(atts1), hash(atts2))
        self.assertEqual(atts1, atts2)

    def test6_SameAttributesOrderMultivalued(self):
        atts1 = Attributes()
        eComs1 = ExtendedCommunities()
        eComs1.communities.append(RouteTarget(64512, 1))
        eComs1.communities.append(Encapsulation(Encapsulation.Type.VXLAN))
        eComs1.communities.append(RouteTarget(64512, 2))
        atts1.add(eComs1)

        atts2 = Attributes()
        eComs2 = ExtendedCommunities()
        eComs2.communities.append(RouteTarget(64512, 2))
        eComs2.communities.append(RouteTarget(64512, 1))
        eComs2.communities.append(Encapsulation(Encapsulation.Type.VXLAN))
        atts2.add(eComs2)

        self.assertEqual(hash(atts1), hash(atts2))
        self.assertEqual(atts1, atts2)

    def test10_Ecoms(self):
        eComs1 = ExtendedCommunities()
        eComs1.communities.append(Encapsulation(Encapsulation.Type.VXLAN))
        atts1 = Attributes()
        atts1.add(eComs1)

        eComs2 = ExtendedCommunities()
        eComs2.communities.append(Encapsulation(Encapsulation.Type.VXLAN))
        eComs2.communities.append(RouteTarget(64512, 1))
        atts2 = Attributes()
        atts2.add(eComs2)

        self.assertFalse(atts1.sameValuesAs(atts2))
        self.assertFalse(atts2.sameValuesAs(atts1))

    def test11_RTs(self):
        rt1a = RouteTarget(64512, 1)
        rt1b = RouteTarget(64512, 1)
        rt3 = RouteTarget(64512, 2)
        rt4 = RouteTarget(64513, 1)

        self.assertEqual(hash(rt1a), hash(rt1b))
        self.assertNotEqual(hash(rt1a), hash(rt3))
        self.assertNotEqual(hash(rt1a), hash(rt4))

        self.assertEqual(rt1a, rt1b)
        self.assertNotEqual(rt1a, rt3)
        self.assertNotEqual(rt1a, rt4)

        self.assertEqual(set([rt1a]), set([rt1b]))
        self.assertEqual(1, len(set([rt1a]).intersection(set([rt1b]))))

    def test12_RTRecord(self):
        rt = RouteTarget(64512, 22)
        rt_record = RTRecord.from_rt(rt)

if __name__ == '__main__':
    unittest.main()



# encoding: utf-8
"""
message.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import sys
from struct import pack

from exabgp.util import character
from exabgp.util import ordinal
from exabgp.util import concat_bytes


class _MessageCode (int):
	if sys.version_info[0] < 3:
		__slots__ = ['SHORT','NAME']

	NOP           = 0x00  # .   0 - internal
	OPEN          = 0x01  # .   1
	UPDATE        = 0x02  # .   2
	NOTIFICATION  = 0x03  # .   3
	KEEPALIVE     = 0x04  # .   4
	ROUTE_REFRESH = 0x05  # .   5
	OPERATIONAL   = 0x06  # .   6  # Not IANA assigned yet

	names = {
		None:           'INVALID',
		NOP:            'NOP',
		OPEN:           'OPEN',
		UPDATE:         'UPDATE',
		NOTIFICATION:   'NOTIFICATION',
		KEEPALIVE:      'KEEPALIVE',
		ROUTE_REFRESH:  'ROUTE_REFRESH',
		OPERATIONAL:    'OPERATIONAL',
	}

	short_names = {
		None:           'invalid',
		NOP:            'nop',
		OPEN:           'open',
		UPDATE:         'update',
		NOTIFICATION:   'notification',
		KEEPALIVE:      'keepalive',
		ROUTE_REFRESH:  'refresh',
		OPERATIONAL:    'operational',
	}

	long_names = {
		None:           'invalid',
		NOP:            'nop',
		OPEN:           'open',
		UPDATE:         'update',
		NOTIFICATION:   'notification',
		KEEPALIVE:      'keepalive',
		ROUTE_REFRESH:  'route refresh',
		OPERATIONAL:    'operational',
	}

	# to_short_names = dict((name,code) for (code,name) in short_names.items())

	def __init__ (self, value):
		self.SHORT = self.short()
		self.NAME = str(self)

	def __str__ (self):
		return self.names.get(self,'unknown message %s' % hex(self))

	def __repr__ (self):
		return str(self)

	def short (self):
		return self.short_names.get(self,'unknown')

# ================================================================== BGP Message
#

# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                                                               |
# +                                                               +
# |                                                               |
# +                                                               +
# |                           Marker                              |
# +                                                               +
# |                                                               |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |          Length               |      Type     |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


class Message (Exception):
	# we need to define TYPE inside __init__ of the subclasses
	# otherwise we can not dynamically create different UnknownMessage
	# TYPE = None

	MARKER = character(0xff)*16
	HEADER_LEN = 19

	registered_message = {}
	# This is redefined by the Notify class, Exception is never used
	klass_notify = Exception
	klass_unknown = Exception

	class CODE (object):
		NOP           = _MessageCode(_MessageCode.NOP)
		OPEN          = _MessageCode(_MessageCode.OPEN)
		UPDATE        = _MessageCode(_MessageCode.UPDATE)
		NOTIFICATION  = _MessageCode(_MessageCode.NOTIFICATION)
		KEEPALIVE     = _MessageCode(_MessageCode.KEEPALIVE)
		ROUTE_REFRESH = _MessageCode(_MessageCode.ROUTE_REFRESH)
		OPERATIONAL   = _MessageCode(_MessageCode.OPERATIONAL)

		MESSAGES = [
			NOP,
			OPEN,
			UPDATE,
			NOTIFICATION,
			KEEPALIVE,
			ROUTE_REFRESH,
			OPERATIONAL
		]

		@staticmethod
		def name (message_id):
			return _MessageCode.names.get(message_id,'unknown message %s' % hex(message_id))

		@staticmethod
		def short (message_id):
			return _MessageCode.short_names.get(message_id,'unknown message %s' % hex(message_id))

		# # Can raise KeyError
		# @staticmethod
		# def code (short):
		# 	return _MessageCode.names.get[short]

		def __init__ (self):
			raise RuntimeError('This class can not be instantiated')

	Length = {
		CODE.OPEN:           lambda _:  _ >= 29,  # noqa
		CODE.UPDATE:         lambda _:  _ >= 23,  # noqa
		CODE.NOTIFICATION:   lambda _:  _ >= 21,  # noqa
		CODE.KEEPALIVE:      lambda _:  _ == 19,  # noqa
		CODE.ROUTE_REFRESH:  lambda _:  _ == 23,  # noqa
	}

	def __init__ (self):
		Exception.__init__(self)

	@staticmethod
	def string (code):
		return _MessageCode.long_names.get(code,'unknown')

	def _message (self, message):
		message_len = pack('!H',19+len(message))
		return concat_bytes(self.MARKER,message_len,self.TYPE,message)

	def message (self,negotiated=None):
		raise NotImplementedError('message not implemented in subclasses')

	@classmethod
	def register (cls, klass):
		if klass.TYPE in cls.registered_message:
			raise RuntimeError('only one class can be registered per message')
		cls.registered_message[ordinal(klass.TYPE)] = klass
		return klass

	@classmethod
	def notify (cls,klass):
		cls.klass_notify = klass
		return klass

	@classmethod
	def klass (cls, what):
		if what in cls.registered_message:
			return cls.registered_message[what]
		raise cls.klass_notify(2,4,'can not handle message %s' % what)

	@classmethod
	def unpack (cls, message, data, negotiated):
		if message in cls.registered_message:
			return cls.klass(message).unpack_message(data,negotiated)
		return cls.klass_unknown(message,data,negotiated)

	@classmethod
	def code (cls,name):
		for message in cls.CODE.MESSAGES:
			if name == str(message) or name == message.short():
				return message
		return cls.CODE.NOP



#!/usr/bin/env python

import sys
import json
import pprint

exit = False

while True:
	line = sys.stdin.readline().strip()

	if not line:
		if exit:
			sys.exit(0)
		exit = True
	else:
		exit = False
		

	print >> sys.stderr, '\n=====\n'
	print >> sys.stderr, line
	print >> sys.stderr, '\n---\n'
	print >> sys.stderr, pprint.pformat(json.loads(line),indent=3).replace("u'","'")
	sys.stderr.flush()



# encoding: utf-8
"""
nodename.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from struct import pack
from struct import unpack

from exabgp.bgp.message.notification import Notify

from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE

#
#     0                   1                   2                   3
#     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    |              Type             |             Length            |
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    //                     Node Name (variable)                    //
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    https://tools.ietf.org/html/rfc7752#section-3.3.1.5 Opaque Node Attribute Format
#
#	  This TLV is added here for completeness but we don't look into the TLV.
#   Use of draft-tantsura-bgp-ls-segment-routing-msd-02 in this TLV is not clear

@LINKSTATE.register()
class NodeOpaque(object):
	TLV = 1025

	def __init__ (self, opaque):
		self.opaque = opaque

	def __repr__ (self):
		return "Node Opaque attribute: %s" % (self.opaque)

	@classmethod
	def unpack (cls,data,length):
		opaque = unpack("!%ds" % length, data[:length])[0]
		return cls(opaque=opaque)



# encoding: utf-8
"""
inet/parser.py

Created by Thomas Mangin on 2015-06-04.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack

from exabgp.util import character
from exabgp.util import concat_bytes_i

from exabgp.bgp.message.update.nlri.qualifier import Labels
from exabgp.bgp.message.update.nlri.qualifier import RouteDistinguisher
from exabgp.bgp.message.update.attribute.sr.prefixsid import PrefixSid
from exabgp.bgp.message.update.attribute.sr.labelindex import SrLabelIndex
from exabgp.bgp.message.update.attribute.sr.srgb import SrGb


def label (tokeniser):
	labels = []
	value = tokeniser()

	if value == '[':
		while True:
			value = tokeniser()
			if value == ']':
				break
			labels.append(int(value))
	else:
		labels.append(int(value))

	return Labels(labels)


def route_distinguisher (tokeniser):
	data = tokeniser()

	separator = data.find(':')
	if separator > 0:
		prefix = data[:separator]
		suffix = int(data[separator+1:])

	if '.' in prefix:
		data = [character(0),character(1)]
		data.extend([character(int(_)) for _ in prefix.split('.')])
		data.extend([character(suffix >> 8),character(suffix & 0xFF)])
		rtd = concat_bytes_i(data)
	else:
		number = int(prefix)
		if number < pow(2,16) and suffix < pow(2,32):
			rtd = character(0) + character(0) + pack('!H',number) + pack('!L',suffix)
		elif number < pow(2,32) and suffix < pow(2,16):
			rtd = character(0) + character(2) + pack('!L',number) + pack('!H',suffix)
		else:
			raise ValueError('invalid route-distinguisher %s' % data)

	return RouteDistinguisher(rtd)

# [ 300, [ ( 800000,100 ), ( 1000000,5000 ) ] ]
def prefix_sid (tokeniser):
	sr_attrs = []
	srgbs = []
	srgb_data = []
	value = tokeniser()
	get_range = False
	consume_extra = False
	try:
		if value == '[':
			label_sid = tokeniser()
			while True:
				value = tokeniser()
				if value == '[':
					consume_extra = True
					continue
				if value == ',':
					continue
				if value == '(':
					while True:
						value = tokeniser()
						if value == ')':
							break
						if value == ',':
							get_range = True
							continue
						if get_range:
							srange = value
							get_range = False
						else:
							base = value
				if value == ')':
					srgb_data.append((base,srange))
					continue
				if value == ']':
					break
		if consume_extra:
			tokeniser()
	except Exception as e:
		raise ValueError('could not parse BGP PrefixSid attribute: {}'.format(e))

	if int(label_sid) < pow(2,32):
		sr_attrs.append(SrLabelIndex(int(label_sid)))

	for srgb in srgb_data:
		if (len(srgb) == 2 and int(srgb[0]) < pow(2,24) and int(srgb[1]) < pow(2,24)):
			srgbs.append((int(srgb[0]),int(srgb[1])))
		else:
			raise ValueError('could not parse SRGB tupple')

	if srgbs:
		sr_attrs.append(SrGb(srgbs))

	return PrefixSid(sr_attrs)



# encoding: utf-8
"""
mprnlri.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import unpack

from exabgp.vendoring import six
from exabgp.util import concat_bytes
from exabgp.util import concat_bytes_i

from exabgp.protocol.ip import NoNextHop
from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.protocol.family import Family

from exabgp.util import character
from exabgp.util import ordinal
from exabgp.bgp.message.direction import IN
# from exabgp.bgp.message.update.attribute.attribute import Attribute
from exabgp.bgp.message.update.attribute import Attribute
from exabgp.bgp.message.update.attribute import NextHop
from exabgp.bgp.message.update.nlri import NLRI

from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.open.capability import Negotiated


# ==================================================== MP Unreacheable NLRI (15)
#

@Attribute.register()
class MPRNLRI (Attribute,Family):
	FLAG = Attribute.Flag.OPTIONAL
	ID = Attribute.CODE.MP_REACH_NLRI

	# __slots__ = ['nlris']

	def __init__ (self, afi, safi, nlris):
		Family.__init__(self,afi,safi)
		# all the routes must have the same next-hop
		self.nlris = nlris

	def __eq__ (self, other):
		return \
			self.ID == other.ID and \
			self.FLAG == other.FLAG and \
			self.nlris == other.nlris

	def __ne__ (self, other):
		return not self.__eq__(other)

	def packed_attributes (self, negotiated, maximum=Negotiated.FREE_SIZE):
		if not self.nlris:
			return

		# addpath = negotiated.addpath.send(self.afi,self.safi)
		# nexthopself = negotiated.nexthopself(self.afi)
		mpnlri = {}
		for nlri in self.nlris:
			if nlri.family() != self.family():  # nlri is not part of specified family
				continue
			if nlri.nexthop is NoNextHop:
				# EOR and Flow may not have any next_hop
				nexthop = b''
			else:
				_,rd_size = Family.size.get(self.family(),(0,0))
				nh_rd = character(0)*rd_size if rd_size else b''
				nexthop = nh_rd + nlri.nexthop.ton(negotiated,nlri.afi)

			# mpunli[nexthop] = nlri
			mpnlri.setdefault(nexthop,[]).append(nlri.pack(negotiated))

		for nexthop,nlris in six.iteritems(mpnlri):
			payload = concat_bytes(self.afi.pack(), self.safi.pack(), character(len(nexthop)), nexthop, character(0))
			header_length = len(payload)
			for nlri in nlris:
				if self._len(payload + nlri) > maximum:
					if len(payload) == header_length or len(payload) > maximum:
						raise Notify(6, 0, 'attributes size is so large we can not even pack on MPRNLRI')
					yield self._attribute(payload)
					payload = concat_bytes(self.afi.pack(), self.safi.pack(), character(len(nexthop)), nexthop, character(0), nlri)
					continue
				payload  = concat_bytes(payload, nlri)
			if len(payload) == header_length or len(payload) > maximum:
				raise Notify(6, 0, 'attributes size is so large we can not even pack on MPRNLRI')
			yield self._attribute(payload)

	def pack (self, negotiated):
		return concat_bytes_i(self.packed_attributes(negotiated))

	def __len__ (self):
		raise RuntimeError('we can not give you the size of an MPRNLRI - was it with our witout addpath ?')
		# return len(self.pack(False))

	def __repr__ (self):
		return "MP_REACH_NLRI for %s %s with %d NLRI(s)" % (self.afi,self.safi,len(self.nlris))

	@classmethod
	def unpack (cls, data, negotiated):
		nlris = []

		# -- Reading AFI/SAFI
		_afi,_safi = unpack('!HB',data[:3])
		afi,safi = AFI.create(_afi),SAFI.create(_safi)
		offset = 3

		# we do not want to accept unknown families
		if negotiated and (afi,safi) not in negotiated.families:
			raise Notify(3,0,'presented a non-negotiated family %s/%s' % (afi,safi))

		# -- Reading length of next-hop
		len_nh = ordinal(data[offset])
		offset += 1

		if (afi,safi) not in Family.size:
			raise Notify(3,0,'unsupported %s %s' % (afi,safi))

		length,rd = Family.size[(afi,safi)]

		if len_nh not in length:
			raise Notify(3,0,'invalid %s %s next-hop length %d expected %s' % (afi,safi,len_nh,' or '.join(str(_) for _ in length)))

		size = len_nh - rd

		# XXX: FIXME: GET IT FROM CACHE HERE ?
		nhs = data[offset+rd:offset+rd+size]
		nexthops = [nhs[pos:pos+16] for pos in range(0,len(nhs),16)]

		# chech the RD is well zero
		if rd and sum([int(ordinal(_)) for _ in data[offset:8]]) != 0:
			raise Notify(3,0,"MP_REACH_NLRI next-hop's route-distinguisher must be zero")

		offset += len_nh

		# Skip a reserved bit as somone had to bug us !
		reserved = ordinal(data[offset])
		offset += 1

		if reserved != 0:
			raise Notify(3,0,'the reserved bit of MP_REACH_NLRI is not zero')

		# Is the peer going to send us some Path Information with the route (AddPath)
		addpath = negotiated.addpath.receive(afi,safi)

		# Reading the NLRIs
		data = data[offset:]

		if not data:
			raise Notify(3,0,'No data to decode in an MPREACHNLRI but it is not an EOR %d/%d' % (afi,safi))

		while data:
			if nexthops:
				for nexthop in nexthops:
					nlri,left = NLRI.unpack_nlri(afi,safi,data,IN.ANNOUNCED,addpath)
					nlri.nexthop = NextHop.unpack(nexthop)
					nlris.append(nlri)
			else:
				nlri,left = NLRI.unpack_nlri(afi,safi,data,IN.ANNOUNCED,addpath)
				nlris.append(nlri)

			if left == data:
				raise RuntimeError("sub-calls should consume data")

			data = left
		return cls(afi,safi,nlris)


EMPTY_MPRNLRI  = MPRNLRI(AFI.undefined,SAFI.undefined,[])



# encoding: utf-8
"""
srrid.py

Created by Evelio Vila
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from exabgp.protocol.ip import IP
from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE

#    draft-gredler-idr-bgp-ls-segment-routing-ext-03
#    0                   1                   2                   3
#    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |            Type               |            Length             |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   //                  IPv4/IPv6 Address (Router-ID)              //
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     Source Router Identifier (Source Router-ID) TLV

@LINKSTATE.register()
class SrSourceRouterID(object):
	TLV = 1171

	def __init__ (self, srid):
		self.srid = srid

	def __repr__ (self):
		return "Source router identifier: %s" % (self.srid)

	@classmethod
	def unpack (cls,data,length):
		if len(data) == 4:
			srid = IP.unpack(data[:4])
		elif len(data) == 16:
			srid = IP.unpack(data[:16])
		return cls(srid=srid)

	def json (self,compact=None):
		return '"sr-source-router-id": "%s"' % str(self.srid)




# encoding: utf-8
"""
__init__.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.util import character

from exabgp.protocol.resource import Resource


# ===================================================================== Protocol
# https://www.iana.org/assignments/protocol-numbers/

class Protocol (Resource):
	NAME = 'protocol'

	ICMP  = 0x01
	IGMP  = 0x02
	TCP   = 0x06
	EGP   = 0x08
	UDP   = 0x11
	RSVP  = 0x2E
	GRE   = 0x2F
	ESP   = 0x32
	AH    = 0x33
	OSPF  = 0x59
	IPIP  = 0x5E
	PIM   = 0x67
	SCTP  = 0x84

	codes = dict ((k.lower().replace('_','-'),v) for (k,v) in {
		'ICMP': ICMP,
		'IGMP': IGMP,
		'TCP':  TCP,
		'EGP':  EGP,
		'UDP':  UDP,
		'RSVP': RSVP,
		'GRE':  GRE,
		'ESP':  ESP,
		'AH':   AH,
		'OSPF': OSPF,
		'IPIP': IPIP,
		'PIM':  PIM,
		'SCTP': SCTP,
	}.items())

	names = dict([(r,l) for (l,r) in codes.items()])

	def pack (self):
		return character(self)



# encoding: utf-8
"""
srcap.py

Created by Evelio Vila
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

import json
from struct import unpack

from exabgp.vendoring.bitstring import BitArray
from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE, LsGenericFlags
from exabgp.bgp.message.notification import Notify

#    draft-gredler-idr-bgp-ls-segment-routing-ext-03
#   0                   1                   2                   3
#    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |               Type            |               Length          |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |      Flags    |   RESERVED    |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   |                  Range Size                   |
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#   //                SID/Label Sub-TLV (variable)                 //
#   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#
#         SR Node Cap Flags
#				+
#      One or more entries, each of which have the following format:
#
#         Range Size: 3 octet value indicating the number of labels in
#         the range.
#
#         SID/Label sub-TLV (as defined in Section 2.3.7.2).

@LINKSTATE.register()
class SrCapabilities(object):
	TLV = 1034

	def __init__ (self, sr_flags, sids):
		self.sr_flags = sr_flags
		self.sids = sids

	def __repr__ (self):
		return "sr_capability_flags: %s, sids: %s" % (self.sr_flags, self.sids)

	@classmethod
	def unpack (cls,data,length):
		# Extract node capability flags
		flags = LsGenericFlags.unpack(data[0:1],LsGenericFlags.ISIS_SR_CAP_FLAGS)
		# Move pointer past flags and reserved bytes
		data = data[2:]
		sids = []
		while data:
			# Range Size: 3 octet value indicating the number of labels in
			# the range.
			b = BitArray(bytes=data[:3])
			range_size = b.unpack('uintbe:24')[0]
			# SID/Label: If length is set to 3, then the 20 rightmost bits
			# represent a label.  If length is set to 4, then the value
			# represents a 32 bit SID.
			t, l = unpack('!HH',data[3:7])
			if t != 1161:
				raise Notify(3,5, "Invalid sub-TLV type: {}".format(t))
			v = data[7:l+7]
			if l == 3:
				b = BitArray(bytes=v)
				sid = b.unpack('uintbe:24')[0]
			elif l == 4:
				sid = unpack('!I',v)[0]
			sids.append((range_size, sid))
			data = data[l+7:]

		return cls(sr_flags=flags, sids=sids)

	def json (self,compact=None):
		return ', '.join(['"sr-capability-flags": {}'.format(self.sr_flags.json()),
			'"sids": {}'.format(json.dumps(self.sids))])



# encoding: utf-8
"""
bandwidth.py

Created by Thomas Mangin on 2017-07-02.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack

from exabgp.bgp.message.update.attribute.community.extended import ExtendedCommunity

# ==================================================================== Bandwidth
# draft-ietf-idr-link-bandwidth-06


@ExtendedCommunity.register
class Bandwidth (ExtendedCommunity):
	COMMUNITY_TYPE = 0x40
	COMMUNITY_SUBTYPE = 0x04

	__slots__ = ['encaps','control','mtu','reserved']

	def __init__ (self, asn, speed, community=None):
		self.asn = asn
		self.speed = speed
		ExtendedCommunity.__init__(
			self,
			community if community is not None else pack(
				"!Hf",asn,speed
			)
		)

	def __repr__ (self):
		return "bandwith:%d:%0.f" % (self.asn,self.speed)

	@staticmethod
	def unpack (data):
		asn, speed = unpack('!Hf',data[2:8])
		return Bandwidth(asn,speed,data[:8])



from exabgp.util import character
from exabgp.util import concat_bytes
from exabgp.bgp.message.update.attribute.community.extended import ExtendedCommunity
from exabgp.bgp.message.update.attribute.community.extended import rt

# draft-ietf-bess-service-chaining


class RTRecord(rt.RouteTarget):
	COMMUNITY_SUBTYPE = 0x13
	DESCRIPTION = "rtrecord"

	@classmethod
	def from_rt(cls, route_target):
		packed = route_target.pack()
		return cls.unpack(concat_bytes(packed[0:1], character(cls.COMMUNITY_SUBTYPE), packed[2:]))


@ExtendedCommunity.register
class RTRecordASN2Number(RTRecord, rt.RouteTargetASN2Number):
	pass


@ExtendedCommunity.register
class RTRecordIPNumber(RTRecord, rt.RouteTargetIPNumber):
	pass


@ExtendedCommunity.register
class RTRecordASN4Number(RTRecord, rt.RouteTargetASN4Number):
	pass



# encoding: utf-8
"""
sr/prefixsid.py

Created by Evelio Vila 2017-02-16
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
"""

import binascii
from struct import unpack
from exabgp.vendoring import six

from exabgp.util import concat_bytes_i
from exabgp.bgp.message.update.attribute.attribute import Attribute

# =====================================================================
# draft-ietf-idr-bgp-prefix-sid
# This Attribute may contain up to 3 TLVs
# Label-Index TLV ( type = 1 ) is mandatory for this attribute.


@Attribute.register()
class PrefixSid (Attribute):
	ID = Attribute.CODE.BGP_PREFIX_SID
	FLAG = FLAG = Attribute.Flag.TRANSITIVE | Attribute.Flag.OPTIONAL
	CACHING = True
	TLV = -1

	# Registered subclasses we know how to decode
	registered_srids = dict()

	def __init__(self, sr_attrs, packed=None):
		self.sr_attrs = sr_attrs
		self._packed = self._attribute(packed if packed else concat_bytes_i(_.pack() for _ in sr_attrs))

	@classmethod
	def register(cls,srid=None,flag=None):
		def register_srid (klass):
			scode = klass.TLV if srid is None else srid
			if scode in cls.registered_srids:
				raise RuntimeError('only one class can be registered per Segment Routing TLV type')
			cls.registered_srids[scode] = klass
			return klass
		return register_srid

	@classmethod
	def unpack(cls, data, negotiated):
		sr_attrs = []
		while data:
			# Type = 1 octet
			scode = six.indexbytes(data,0)
			# L = 2 octet  :|
			length = unpack('!H',data[1:3])[0]
			if scode in cls.registered_srids:
				klass = cls.registered_srids[scode].unpack(data[3:length+3],length)
			else:
				klass = GenericSRId(scode,data[3:length+3])
			klass.TLV = scode
			sr_attrs.append(klass)
			data = data[length+3:]
		return cls(sr_attrs=sr_attrs)

	def json (self,compact=None):
		content = ', '.join(d.json() for d in self.sr_attrs)
		return '{ %s }' % (content)

	def __str__(self):
		label_index = next((i for i in self.sr_attrs if i.TLV == 1), None)
		srgb = next((i for i in self.sr_attrs if i.TLV == 3), None)
		if srgb:
			return "[ {}, {} ]".format(str(label_index), str(srgb))
		else:
			return "[ {} ]".format(str(label_index))

	def pack (self, negotiated=None):
		return self._packed


class GenericSRId(object):
	TLV = 99998

	def __init__ (self, code, rep):
		self.rep = rep
		self.code = code

	def __repr__ (self):
		return "Attribute with code [ %s ] not implemented" % (self.code)

	@classmethod
	def unpack (cls,scode,data):
		length = len(data)
		info = binascii.b2a_uu(data[:length])
		return cls(code=scode,rep=info)

	def json (self,compact=None):
		return '"attribute-not-implemented": "%s"' % (self.code)



# encoding: utf-8
"""
template.py

Created by Thomas Mangin on 2015-06-16.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.configuration.core import Section
from exabgp.configuration.neighbor import ParseNeighbor


class ParseTemplateNeighbor (Section):
	syntax = \
		'neighbor {\n' \
		'   <neighbor commands>\n' \
		'}'

	known = ParseNeighbor.known
	action = ParseNeighbor.action
	default = ParseNeighbor.default

	name = 'template-neighbor'

	def __init__ (self, tokeniser, scope, error, logger):
		Section.__init__(self,tokeniser,scope,error,logger)

	def clear (self):
		self._names = []

	def pre (self):
		named = self.tokeniser.line[1]
		self.check_name(named)
		self.scope.enter(named)
		self.scope.to_context()
		return True

	def post (self):
		routes = self.scope.pop_routes()
		self.scope.extend('routes',routes)
		self.scope.leave()
		return True



# encoding: utf-8
"""
inet/__init__.py

Created by Thomas Mangin on 2015-06-04.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.configuration.static.route import ParseStaticRoute
from exabgp.configuration.static.parser import prefix

from exabgp.protocol.ip import IP
from exabgp.protocol.family import SAFI

from exabgp.bgp.message import OUT
from exabgp.bgp.message.update.nlri import CIDR
from exabgp.bgp.message.update.nlri import INET
from exabgp.bgp.message.update.nlri import Label
from exabgp.bgp.message.update.nlri import IPVPN

from exabgp.bgp.message.update.attribute import Attributes

from exabgp.rib.change import Change

from exabgp.configuration.static.mpls import label
from exabgp.configuration.static.mpls import route_distinguisher

class ParseStatic (ParseStaticRoute):
	syntax = \
		'route <ip>/<netmask> %s;' % ' '.join(ParseStaticRoute.definition)

	action = dict(ParseStaticRoute.action)

	name = 'static'

	def __init__ (self, tokeniser, scope, error, logger):
		ParseStaticRoute.__init__(self,tokeniser,scope,error,logger)

	def clear (self):
		return True

	def pre (self):
		return True

	def post (self):
		return True


@ParseStatic.register('route','append-route')
def route (tokeniser):
	ipmask = prefix(tokeniser)

	if 'rd' in tokeniser.tokens or 'route-distinguisher' in tokeniser.tokens:
		nlri = IPVPN(IP.toafi(ipmask.top()),SAFI.mpls_vpn,OUT.ANNOUNCE)
	elif 'label' in tokeniser.tokens:
		nlri = Label(IP.toafi(ipmask.top()),SAFI.nlri_mpls,OUT.ANNOUNCE)
	else:
		nlri = INET(IP.toafi(ipmask.top()),IP.tosafi(ipmask.top()),OUT.ANNOUNCE)

	nlri.cidr = CIDR(ipmask.pack(),ipmask.mask)

	change = Change(
		nlri,
		Attributes()
	)

	while True:
		command = tokeniser()

		if not command:
			break

		if command == 'label':
			nlri.labels = label(tokeniser)
			continue

		if command == 'rd' or command == 'route-distinguisher':
			nlri.rd = route_distinguisher(tokeniser)
			continue

		action = ParseStatic.action.get(command,'')

		if action == 'attribute-add':
			change.attributes.add(ParseStatic.known[command](tokeniser))
		elif action == 'nlri-set':
			change.nlri.assign(ParseStatic.assign[command],ParseStatic.known[command](tokeniser))
		elif action == 'nexthop-and-attribute':
			nexthop,attribute = ParseStatic.known[command](tokeniser)
			change.nlri.nexthop = nexthop
			change.attributes.add(attribute)
		else:
			raise ValueError('route: unknown command "%s"' % command)

	return list(ParseStatic.split(change))


@ParseStatic.register('attributes','append-route')
def attributes (tokeniser):
	ipmask = prefix(lambda: tokeniser.tokens[-1])

	if 'rd' in tokeniser.tokens or 'route-distinguisher' in tokeniser.tokens:
		nlri = IPVPN(IP.toafi(ipmask.top()),SAFI.mpls_vpn,OUT.ANNOUNCE)
	elif 'label' in tokeniser.tokens:
		nlri = Label(IP.toafi(ipmask.top()),SAFI.nlri_mpls,OUT.ANNOUNCE)
	else:
		nlri = INET(IP.toafi(ipmask.top()),IP.tosafi(ipmask.top()),OUT.ANNOUNCE)

	nlri.cidr = CIDR(ipmask.pack(),ipmask.mask)
	attr = Attributes()

	labels = None
	rd = None

	while True:
		command = tokeniser()

		if not command:
			return []

		if command == 'nlri':
			break

		if command == 'label':
			labels = label(tokeniser)
			continue

		if command == 'rd' or command == 'route-distinguisher':
			rd = route_distinguisher(tokeniser)
			continue

		action = ParseStatic.action[command]

		if action == 'attribute-add':
			attr.add(ParseStatic.known[command](tokeniser))
		elif action == 'nlri-set':
			nlri.assign(ParseStatic.assign[command],ParseStatic.known[command](tokeniser))
		elif action == 'nexthop-and-attribute':
			nexthop,attribute = ParseStatic.known[command](tokeniser)
			nlri.nexthop = nexthop
			attr.add(attribute)
		else:
			raise ValueError('route: unknown command "%s"' % command)

	changes = []
	while True:
		peeked_nlri = tokeniser.peek()
		if not peeked_nlri:
			break

		ipmask = prefix(tokeniser)
		new = Change(
			nlri.__class__(
				nlri.afi,
				nlri.safi,
				OUT.UNSET
			),
			attr
		)
		new.nlri.cidr = CIDR(ipmask.pack(),ipmask.mask)
		if labels:
			new.nlri.labels = labels
		if rd:
			new.nlri.rd = rd
		new.nlri.nexthop = nlri.nexthop
		changes.append(new)

	return changes



# encoding: utf-8
"""
igptags.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from struct import unpack

from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE

#   The IGP Route Tag TLV carries original IGP Tags (IS-IS [RFC5130] or
#   OSPF) of the prefix and is encoded as follows:
#
#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |              Type             |             Length            |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     //                    Route Tags (one or more)                 //
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     https://tools.ietf.org/html/rfc7752#section-3.3.3.2


@LINKSTATE.register()
class IgpTags(object):
	TLV = 1153

	def __init__ (self, igptags):
		self.igptags = igptags

	def __repr__ (self):
		return "IGP Route Tags: %s" % (self.igptags)

	@classmethod
	def unpack (cls,data,length):
		tags = []
		n = length // 4
		ind = 0
		for i in list(range(n)):
			tag = unpack("!L", data[ind:4*(i+1)])[0]
			tags.append(tag)
			ind += 4
		return cls(igptags=tags)

	def json (self,compact=None):
		return '"igp-route-tags": %s' % self.igptags



# encoding: utf-8
"""
igpextags.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from struct import unpack

from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE

#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |              Type             |             Length            |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     //                Extended Route Tag (one or more)             //
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     https://tools.ietf.org/html/rfc7752#section-3.3.3.3


@LINKSTATE.register()
class IgpExTags(object):
	TLV = 1154

	def __init__ (self, igpextags):
		self.igpextags = igpextags

	def __repr__ (self):
		return "IGP Extended Route Tags: %s" % (self.igpextags)

	@classmethod
	def unpack (cls,data,length):
		tags = []
		n = length // 8
		ind = 0
		for i in list(range(n)):
			tag = unpack("!Q", data[ind:8*(i+1)])[0]
			tags.append(tag)
			ind += 8
		return cls(igpextags=tags)

	def json (self,compact=None):
		return '"igp-extended-route-tags": %s' % self.igpextags



#!/usr/bin/env python

from __future__ import print_function
import sys
import signal

class TimeError (Exception): pass

def handler (signum, frame):
	raise TimeError()

count = 0

while True:
	try:
		signal.signal(signal.SIGALRM, handler)
		signal.alarm(4)

		line = sys.stdin.readline()
		sys.stderr.write('received %s\n' % line.strip())
		sys.stderr.flush()
	except TimeError:
		print('announce route-refresh ipv4 unicast')
		sys.stdout.flush()
		print('announce route-refresh ipv4 unicast', file=sys.stderr)
		sys.stderr.flush()



#!/usr/bin/env python
# encoding: utf-8
"""
cli.py

Created by Thomas Mangin on 2014-12-22.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import os
import sys
import time
import select
import signal
import errno

from exabgp.application.bgp import root_folder
from exabgp.application.bgp import named_pipe
from exabgp.application.bgp import get_envfile
from exabgp.application.bgp import get_env
from exabgp.application.control import check_fifo

from exabgp.reactor.network.error import error
from exabgp.reactor.api.response.answer import Answer

from exabgp.vendoring import docopt

usage = """\
The BGP swiss army knife of networking

usage: exabgpcli [--root ROOT]
\t\t\t\t\t\t\t\t [--help|<command>...]
\t\t\t\t\t\t\t\t [--env ENV]

positional arguments:
\tcommand               valid exabgpcli command (see below)

optional arguments:
\t--env ENV,   -e ENV   environment configuration file
\t--help,      -h       exabgp manual page
\t--root ROOT, -f ROOT  root folder where etc,bin,sbin are located

commands:
\thelp                  show the commands known by ExaBGP
""".replace('\t','  ')


class AnswerStream:
	done = '\n%s\n' % Answer.done
	error = '\n%s\n' % Answer.error
	shutdown = '\n%s\n' % Answer.error
	buffer_size = Answer.buffer_size + 2

def main ():
	options = docopt.docopt(usage, help=False)
	if options['--env'] is None:
		options['--env'] = ''

	root = root_folder(options,['/bin/exabgpcli','/sbin/exabgpcli','/lib/exabgp/application/cli.py'])
	prefix = '' if root == '/usr' else root
	etc = prefix + '/etc/exabgp'
	envfile = get_envfile(options,etc)
	env = get_env(envfile)
	pipename = env['api']['pipename']

	if options['--help']:
		sys.stdout.write(usage)
		sys.stdout.flush()
		sys.exit(0)

	if not options['<command>']:
		sys.stdout.write(usage)
		sys.stdout.flush()
		sys.exit(0)

	command = ' '.join(options['<command>'])

	pipes = named_pipe(root, pipename)
	if len(pipes) != 1:
		sys.stdout.write('could not find ExaBGP\'s named pipes (%s.in and %s.out) for the cli\n' % (pipename, pipename))
		sys.stdout.write('we scanned the following folders (the number is your PID):\n - ')
		sys.stdout.write('\n - '.join(pipes))
		sys.stdout.flush()
		sys.exit(1)

	send = pipes[0] + pipename + '.in'
	recv = pipes[0] + pipename + '.out'

	if not check_fifo(send):
		sys.stdout.write('could not find write named pipe to connect to ExaBGP')
		sys.stdout.flush()
		sys.exit(1)

	if not check_fifo(recv):
		sys.stdout.write('could not find read named pipe to connect to ExaBGP')
		sys.stdout.flush()
		sys.exit(1)

	def write_timeout(signum, frame):
		sys.stderr.write('could not send command to ExaBGP')
		sys.stderr.flush()
		sys.exit(1)

	buffer = ''
	start = time.time()
	try:
		reader = os.open(recv, os.O_RDONLY | os.O_EXCL | os.O_NONBLOCK)
		while True:
			while select.select([reader], [], [], 0) != ([], [], []):
				buffer += os.read(reader,4096)
				buffer = buffer[-AnswerStream.buffer_size:]
			# we read nothing, nothing to do
			if not buffer:
				break
			# we read some data but it is not ending by a new line (ie: not a command completion)
			if buffer[-1] != '\n':
				continue
			if AnswerStream.done.endswith(buffer[-len(AnswerStream.done):]):
				break
			if AnswerStream.error.endswith(buffer[-len(AnswerStream.error):]):
				break
			if AnswerStream.shutdown.endswith(buffer[-len(AnswerStream.shutdown):]):
				break
			# we are not ack'ing the command and probably have read all there is
			if time.time() > start + 1.5:
				break

	except Exception as exc:
		sys.stdout.write('could not clear named pipe from potential previous command data')
		sys.stdout.write(exc)
		sys.stdout.flush()

	signal.signal(signal.SIGALRM, write_timeout)
	signal.alarm(2)

	try:
		writer = os.open(send, os.O_WRONLY | os.O_EXCL)
		os.write(writer,command.encode('utf-8') + b'\n')
		os.close(writer)
	except OSError as exc:
		if exc.errno == errno.ENXIO:
			sys.stdout.write('ExaBGP is not running / using the configured named pipe')
			sys.stdout.flush()
			sys.exit(1)
		sys.stdout.write('could not communicate with ExaBGP')
		sys.stdout.flush()
		sys.exit(1)
	except IOError as exc:
		sys.stdout.write('could not communicate with ExaBGP')
		sys.stdout.flush()
		sys.exit(1)

	signal.alarm(0)

	if command == 'reset':
		sys.exit(0)

	def read_timeout(signum, frame):
		sys.stderr.write('could not read answer to ExaBGP')
		sys.stderr.flush()
		sys.exit(1)

	signal.signal(signal.SIGALRM, read_timeout)

	try:
		signal.alarm(5)
		reader = os.open(recv, os.O_RDONLY | os.O_EXCL)
		signal.alarm(0)

		buf = b''
		done = False
		while not done:
			try:
				raw = os.read(reader,4096)
				buf += raw
				while b'\n' in buf:
					line,buf = buf.split(b'\n',1)
					string = line.decode()
					if string == Answer.done:
						done = True
						break
					if string == Answer.shutdown:
						sys.stderr.write('ExaBGP is shutting down, command aborted\n')
						sys.stderr.flush()
						done = True
						break
					if string == Answer.error:
						done = True
						sys.stderr.write('ExaBGP returns an error (see ExaBGP\'s logs for more information)\n')
						sys.stderr.write('use help for a list of available commands\n')
						sys.stderr.flush()
						break
					sys.stdout.write('%s\n' % string)
					sys.stdout.flush()

				select.select([reader],[],[],0.01)
			except OSError as exc:
				if exc.errno in error.block:
					break
			except IOError as exc:
				if exc.errno in error.block:
					break
		os.close(reader)

		sys.exit(0)
	except IOError:
		sys.stdout.write('could not read answer from ExaBGP')
		sys.stdout.flush()
	except OSError:
		sys.stdout.write('could not read answer from ExaBGP')
		sys.stdout.flush()


if __name__ == '__main__':
	try:
		main()
	except KeyboardInterrupt:
		pass



# encoding: utf-8
"""
asn4.py

Created by Thomas Mangin on 2014-06-30.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.bgp.message.open.asn import ASN
from exabgp.bgp.message.open.capability.capability import Capability

# ========================================================================= ASN4
#


@Capability.register()
class ExtendedMessage (Capability):
	ID = Capability.CODE.EXTENDED_MESSAGE
	INITIAL_SIZE = 4096
	EXTENDED_SIZE = 65535

	def __str__ (self):
		return 'Extended Message(%d)' % self.EXTENDED_SIZE

	def extract (self):
		return [b'']

	@staticmethod
	def unpack_capability (instance, data, capability=None):  # pylint: disable=W0613
		return ExtendedMessage()

	def json (self):
		return '{ "name": "extended-message", "size": %d }' % self.EXTENDED_SIZE



#!/usr/bin/python
"""
flow.py

Created by Thomas Mangin on 2017-07-06.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

# based on the blog at: http://blog.sflow.com/2017/07/bgp-flowspec-on-white-box-switch.html

import os
import sys
import json
import re
import subprocess
import signal


class ACL (object):
	dry = os.environ.get('CUMULUS_FLOW_RIB',False)

	path = '/etc/cumulus/acl/policy.d/'
	priority = '60'
	prefix = 'flowspec'
	bld = '.bld'
	suffix = '.rules'

	__uid = 0
	_known = dict()

	@classmethod
	def _uid (cls):
		cls.__uid += 1
		return cls.__uid

	@classmethod
	def _file(cls,name):
		return cls.path+cls.priority+cls.prefix+str(name)+cls.suffix

	@classmethod
	def _delete (cls,key):
		if key not in cls._known:
			return
		# removing key first so the call to clear never loops forever
		uid,acl = cls._known.pop(key)
		try:
			filename = cls._file(uid)
			if os.path.isfile(filename):
				os.unlink(filename)
		except KeyboardInterrupt:
			raise
		except Exception:
			pass

	@classmethod
	def _commit(cls):
		if cls.dry:
			cls.show()
			return
		try:
			return subprocess.Popen(
				['cl-acltool','-i'],
				stderr=subprocess.STDOUT,
				stdout=subprocess.PIPE
			).communicate()[0]
		except KeyboardInterrupt:
			raise
		except Exception:
			pass

	@staticmethod
	def _build (flow, action):
		acl = '[iptables]\n-A FORWARD --in-interface swp+'
		if 'protocol' in flow:
			acl += ' -p ' + re.sub('[!<>=]','',flow['protocol'][0])
		if 'source-ipv4' in flow:
			acl += ' -s ' + flow['source-ipv4'][0]
		if 'destination-ipv4' in flow:
			acl += ' -d ' + flow['destination-ipv4'][0]
		if 'source-port' in flow:
			acl += ' --sport ' + re.sub('[!<>=]','',flow['source-port'][0])
		if 'destination-port' in flow:
			acl += ' --dport ' + re.sub('[!<>=]','',flow['destination-port'][0])
		acl = acl + ' -j DROP\n'
		return acl

	@classmethod
	def insert (cls,flow,action):
		key = flow['string']
		if key in cls._known:
			return
		uid = cls._uid()
		acl = cls._build(flow,action)
		cls._known[key] = (uid,acl)
		try:
			with open(cls._file(uid),'w') as f:
				f.write(acl)
			cls._commit()
		except KeyboardInterrupt:
			raise
		except Exception:
			cls.end()

	@classmethod
	def remove (cls,flow):
		key = flow['string']
		if key not in cls._known:
			return
		uid,_ = cls._known[key]
		cls._delete(key)

	@classmethod
	def clear (cls):
		for key in cls._known:
			cls._delete(key)
		cls._commit()

	@classmethod
	def end (cls):
		cls.clear()
		sys.exit(1)

	@classmethod
	def show (cls):
		for key,(uid,_) in cls._known.items():
			sys.stderr.write('%d %s\n' % (uid,key))
		for _,acl in cls._known.values():
			sys.stderr.write('%s' % acl)
		sys.stderr.flush()


signal.signal(signal.SIGTERM,ACL.end)


opened = 0
buffered = ''

while True:
	try:
		line = sys.stdin.readline()
		if not line or 'shutdown' in line:
			ACL.end()
		buffered += line
		opened += line.count('{')
		opened -= line.count('}')
		if opened:
			continue
		line, buffered = buffered, ''
		message = json.loads(line)

		if message['type'] == 'state' and message['neighbor']['state'] == 'down':
			ACL.clear()
			continue

		if message['type'] != 'update':
			continue

		update = message['neighbor']['message']['update']

		if 'announce' in update:
			flow = update['announce']['ipv4 flow']
			# The RFC allows both encoding
			flow = flow['no-nexthop'][0] if 'no-nexthop' in flow else flow[0]

			community = update['attribute']['extended-community'][0]
			ACL.insert(flow,community)
			continue

		if 'withdraw' in update:
			flow = update['withdraw']['ipv4 flow'][0]
			ACL.remove(flow)
			continue

	except KeyboardInterrupt:
		ACL.end()
	except Exception:
		pass



"""
prefix.py

Created by Diego Garcia del Rio on 2015-03-12.
Copyright (c) 2015 Alcatel-Lucent. All rights reserved.

Based on work by Thomas Morin on mac.py
Copyright (c) 2014-2017 Orange. All rights reserved.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.protocol.ip import IP
from exabgp.util import character
from exabgp.util import ordinal
from exabgp.util import concat_bytes
from exabgp.bgp.message.update.nlri.qualifier import RouteDistinguisher
from exabgp.bgp.message.update.nlri.qualifier import Labels
from exabgp.bgp.message.update.nlri.qualifier import ESI
from exabgp.bgp.message.update.nlri.qualifier import EthernetTag

from exabgp.bgp.message.update.nlri import NLRI
from exabgp.bgp.message.update.nlri.evpn.nlri import EVPN

from exabgp.bgp.message.notification import Notify


# ------------ EVPN Prefix Advertisement NLRI ------------
# As described here:
# https://tools.ietf.org/html/draft-ietf-bess-evpn-prefix-advertisement-01

# +---------------------------------------+
# |      RD   (8 octets)                  |
# +---------------------------------------+
# |Ethernet Segment Identifier (10 octets)|
# +---------------------------------------+
# |  Ethernet Tag ID (4 octets)           |
# +---------------------------------------+
# |  IP Prefix Length (1 octet)           |
# +---------------------------------------+
# |  IP Prefix (4 or 16 octets)           |
# +---------------------------------------+
# |  GW IP Address (4 or 16 octets)       |
# +---------------------------------------+
# |  MPLS Label (3 octets)                |
# +---------------------------------------+
# total NLRI length is 34 bytes for IPv4 or 58 bytes for IPv6

# ======================================================================= Prefix

# https://tools.ietf.org/html/draft-rabadan-l2vpn-evpn-prefix-advertisement-03

@EVPN.register
class Prefix (EVPN):
	CODE = 5
	NAME = "IP Prefix Advertisement"
	SHORT_NAME = "PrfxAdv"

	def __init__(self, rd, esi, etag, label, ip, iplen, gwip, packed=None,nexthop=None,action=None,addpath=None):
		'''
		rd: a RouteDistinguisher
		esi: an EthernetSegmentIdentifier
		etag: an EthernetTag
		mac: a MAC
		label: a LabelStackEntry
		ip: an IP address (dotted quad string notation)
		iplen: prefixlength for ip (defaults to 32)
		gwip: an IP address (dotted quad string notation)
		'''
		EVPN.__init__(self,action,addpath)
		self.nexthop = nexthop
		self.rd = rd
		self.esi = esi
		self.etag = etag
		self.ip = ip
		self.iplen = iplen
		self.gwip = gwip
		self.label = label
		self.label = label if label else Labels.NOLABEL
		self._pack(packed)

	def __eq__ (self, other):
		return \
			NLRI.__eq__(self,other) and \
			self.CODE == other.CODE and \
			self.rd == other.rd and \
			self.etag == other.etag and \
			self.ip == other.ip and \
			self.iplen == other.iplen
		# esi, label and gwip must not be compared

	def __ne__ (self, other):
		return not self.__eq__(other)

	def __str__ (self):
		return "%s:%s:%s:%s:%s%s:%s:%s" % (
			self._prefix(),
			self.rd._str(),
			self.esi,
			self.etag,
			self.ip,
			"/%d" % self.iplen,
			self.gwip,
			self.label
		)

	def __hash__ (self):
		# esi, and label, gwip must *not* be part of the hash
		return hash("%s:%s:%s:%s" % (self.rd,self.etag,self.ip,self.iplen))

	def _pack (self, packed=None):
		if self._packed:
			return self._packed

		if packed:
			self._packed = packed
			return packed

		self._packed = concat_bytes(
			self.rd.pack(),
			self.esi.pack(),
			self.etag.pack(),
			character(self.iplen),
			self.ip.pack(),
			self.gwip.pack(),
			self.label.pack(),
		)
		return self._packed

	@classmethod
	def unpack (cls, exdata):
		data = exdata

		# Get the data length to understand if addresses are IPv4 or IPv6
		datalen = len(data)

		rd = RouteDistinguisher.unpack(data[:8])
		data = data[8:]

		esi = ESI.unpack(data[:10])
		data = data[10:]

		etag = EthernetTag.unpack(data[:4])
		data = data[4:]

		iplen = ordinal(data[0])
		data = data[1:]

		if datalen == (26 + 8):  # Using IPv4 addresses
			ip = IP.unpack(data[:4])
			data = data[4:]
			gwip = IP.unpack(data[:4])
			data = data[4:]
		elif datalen == (26 + 32):  # Using IPv6 addresses
			ip = IP.unpack(data[:16])
			data = data[16:]
			gwip = IP.unpack(data[:16])
			data = data[16:]
		else:
			raise Notify(3,5,"Data field length is given as %d, but EVPN route currently support only IPv4 or IPv6(34 or 58)" % datalen)

		label = Labels.unpack(data[:3])

		return cls(rd,esi,etag,label,ip,iplen,gwip,exdata)

	def json (self, compact=None):
		content = ' "code": %d, ' % self.CODE
		content += '"parsed": true, '
		content += '"raw": "%s", ' % self._raw()
		content += '"name": "%s", ' % self.NAME
		content += '%s, ' % self.rd.json()
		content += '%s, ' % self.esi.json()
		content += '%s, ' % self.etag.json()
		content += '%s, ' % self.label.json()
		content += '"ip": "%s", ' % str(self.ip)
		content += '"iplen": %d, ' % self.iplen
		content += '"gateway": "%s" ' % str(self.gwip)
		return '{%s}' % content



# encoding: utf-8
"""
__init__.py

Created by Thomas Mangin on 2015-05-15.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import time
import socket


__warned = False
__host_name = ''
__domain_name = ''

__warning = '''
Your OS is very slow when returning the host FQDN
Most likely you do not have valid forward/reverse DNS setup
Adding your hostname to the /etc/hosts file should fix the issue
'''


def host ():
	global __host_name
	if not __host_name:
		value = socket.gethostname()
		__host_name = value.split('.')[0] if value else 'localhost'
	return __host_name


def domain ():
	global __domain_name
	if not __domain_name:
		value = socket.getfqdn()
		__domain_name = value.split('.')[0] if value else 'localhost'
	return __domain_name


def warn ():
	if __warned:
		return ''

	now = time.time()
	_ = host(),domain()
	if time.time() - now > 1.0:
		return __warning



# encoding: utf-8
"""
protocol.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import os

from exabgp.vendoring import six
import traceback

# ================================================================ Registration
#

from exabgp.util import ordinal

from exabgp.reactor.network.outgoing import Outgoing
# from exabgp.reactor.network.error import NotifyError

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.bgp.message import Message
from exabgp.bgp.message import NOP
from exabgp.bgp.message import _NOP
from exabgp.bgp.message import Open
from exabgp.bgp.message.open import Version
from exabgp.bgp.message.open.capability import Capabilities
from exabgp.bgp.message.open.capability import Negotiated
from exabgp.bgp.message import Update
from exabgp.bgp.message import EOR
from exabgp.bgp.message import KeepAlive
from exabgp.bgp.message import Notification
from exabgp.bgp.message import Notify
from exabgp.bgp.message import Operational

from exabgp.bgp.message.direction import IN
from exabgp.bgp.message.update.attribute import Attribute

from exabgp.protocol.ip import IP
from exabgp.reactor.api.processes import ProcessError

from exabgp.logger import Logger
from exabgp.logger import FakeLogger

# This is the number of chuncked message we are willing to buffer, not the number of routes
MAX_BACKLOG = 15000

_UPDATE = Update([],b'')
_OPERATIONAL = Operational(0x00)


class Protocol (object):
	decode = True

	def __init__ (self, peer):
		try:
			self.logger = Logger()
		except RuntimeError:
			self.logger = FakeLogger()
		self.peer = peer
		self.neighbor = peer.neighbor
		self.negotiated = Negotiated(self.neighbor)
		self.connection = None

		if self.neighbor.connect:
			self.port = self.neighbor.connect
		elif os.environ.get('exabgp.tcp.port','').isdigit():
			self.port = int(os.environ.get('exabgp.tcp.port'))
		elif os.environ.get('exabgp_tcp_port','').isdigit():
			self.port = int(os.environ.get('exabgp_tcp_port'))
		else:
			self.port = 179

		from exabgp.configuration.environment import environment
		self.log_routes = peer.neighbor.adj_rib_in or environment.settings().log.routes

	def fd (self):
		if self.connection is None:
			return None
		return self.connection.fd()

	# XXX: we use self.peer.neighbor.peer_address when we could use self.neighbor.peer_address

	def me (self, message):
		return "%s/%s %s" % (self.peer.neighbor.peer_address,self.peer.neighbor.peer_as,message)

	def accept (self, incoming):
		self.connection = incoming

		if self.peer.neighbor.api['neighbor-changes']:
			self.peer.reactor.processes.connected(self.peer.neighbor)

		# very important - as we use this function on __init__
		return self

	def connect (self):
		# allows to test the protocol code using modified StringIO with a extra 'pending' function
		if not self.connection:
			local = self.neighbor.md5_ip.top() if not self.neighbor.auto_discovery else None
			peer = self.neighbor.peer_address.top()
			afi = self.neighbor.peer_address.afi
			md5 = self.neighbor.md5_password
			md5_base64 = self.neighbor.md5_base64
			ttl_out = self.neighbor.ttl_out
			self.connection = Outgoing(afi,peer,local,self.port,md5,md5_base64,ttl_out)
			if not self.connection.init:
				yield False
				return
			if not local:
				self.neighbor.local_address = IP.create(self.connection.local)
				if self.neighbor.router_id is None and self.neighbor.local_address.afi == AFI.ipv4:
					self.neighbor.router_id = self.neighbor.local_address

			for connected in self.connection.establish():
				if not connected:
					yield False
					continue
				if self.peer.neighbor.api['neighbor-changes']:
					self.peer.reactor.processes.connected(self.peer.neighbor)
				yield True
				return


	def close (self, reason='protocol closed, reason unspecified'):
		if self.connection:
			self.logger.debug(reason,self.connection.session())

			# must be first otherwise we could have a loop caused by the raise in the below
			self.connection.close()
			self.connection = None

			self.peer.stats['down'] = self.peer.stats.get('down',0) + 1
			try:
				if self.peer.neighbor.api['neighbor-changes']:
					self.peer.reactor.processes.down(self.peer.neighbor,reason)
			except ProcessError:
				self.logger.debug('could not send notification of neighbor close to API',self.connection.session())

	def _to_api (self,direction,message,raw):
		packets = self.neighbor.api['%s-packets' % direction]
		parsed = self.neighbor.api['%s-parsed' % direction]
		consolidate = self.neighbor.api['%s-consolidate' % direction]
		negotiated = self.negotiated if self.neighbor.api['negotiated'] else None

		if consolidate:
			if packets:
				self.peer.reactor.processes.message(self.peer.neighbor,direction,message,negotiated,raw[:19],raw[19:])
			else:
				self.peer.reactor.processes.message(self.peer.neighbor,direction,message,negotiated,b'',b'')
		else:
			if packets:
				self.peer.reactor.processes.packets(self.peer.neighbor,direction,int(message.ID),negotiated,raw[:19],raw[19:])
			if parsed:
				self.peer.reactor.processes.message(message.ID,self.peer.neighbor,direction,message,negotiated,b'',b'')

	def write (self, message, negotiated=None):
		raw = message.message(negotiated)

		code = 'send-%s' % Message.CODE.short(message.ID)
		self.peer.stats[code] = self.peer.stats.get(code,0) + 1
		if self.neighbor.api.get(code,False):
			self._to_api('send',message,raw)

		for boolean in self.connection.writer(raw):
			yield boolean

	def send (self, raw):
		code = 'send-%s' % Message.CODE.short(ordinal(raw[18]))
		self.peer.stats[code] = self.peer.stats.get(code,0) + 1
		if self.neighbor.api.get(code,False):
			message = Update.unpack_message(raw[19:],self.negotiated)
			self._to_api('send',message,raw)

		for boolean in self.connection.writer(raw):
			yield boolean

	# Read from network .......................................................

	def read_message (self):
		# This will always be defined by the loop but scope leaking upset scrutinizer/pylint
		msg_id = None

		packets = self.neighbor.api['receive-packets']
		consolidate = self.neighbor.api['receive-consolidate']
		parsed = self.neighbor.api['receive-parsed']

		body,header = b'',b''  # just because pylint/pylama are getting more clever

		for length,msg_id,header,body,notify in self.connection.reader():
			# internal issue
			if notify:
				code = 'receive-%s' % Message.CODE.NOTIFICATION.SHORT
				if self.neighbor.api.get(code,False):
					if consolidate:
						self.peer.reactor.processes.notification(self.peer.neighbor,'receive',notify.code,notify.subcode,str(notify),None,header,body)
					elif parsed:
						self.peer.reactor.processes.notification(self.peer.neighbor,'receive',notify.code,notify.subcode,str(notify),None,b'',b'')
					elif packets:
						self.peer.reactor.processes.packets(self.peer.neighbor,'receive',msg_id,None,header,body)
				# XXX: is notify not already Notify class ?
				raise Notify(notify.code,notify.subcode,str(notify))

			if not length:
				yield _NOP
				continue

			self.logger.debug('<< message of type %s' % Message.CODE.name(msg_id),self.connection.session())

			code = 'receive-%s' % Message.CODE.short(msg_id)
			self.peer.stats[code] = self.peer.stats.get(code,0) + 1
			for_api = self.neighbor.api.get(code,False)

			if for_api and packets and not consolidate:
				negotiated = self.negotiated if self.neighbor.api.get('negotiated',False) else None
				self.peer.reactor.processes.packets(self.peer.neighbor,'receive',msg_id,negotiated,header,body)

			if msg_id == Message.CODE.UPDATE:
				if not self.neighbor.adj_rib_in and not (for_api or self.log_routes) and not (parsed or consolidate):
					yield _UPDATE
					return

			try:
				message = Message.unpack(msg_id,body,self.negotiated)
			except (KeyboardInterrupt,SystemExit,Notify):
				raise
			except Exception as exc:
				self.logger.debug('could not decode message "%d"' % msg_id,self.connection.session())
				self.logger.debug('%s' % str(exc),self.connection.session())
				self.logger.debug(traceback.format_exc(),self.connection.session())
				raise Notify(1,0,'can not decode update message of type "%d"' % msg_id)
				# raise Notify(5,0,'unknown message received')

			if message.TYPE == Update.TYPE:
				if Attribute.CODE.INTERNAL_TREAT_AS_WITHDRAW in message.attributes:
					for nlri in message.nlris:
						nlri.action = IN.WITHDRAWN

			if for_api:
				negotiated = self.negotiated if self.neighbor.api.get('negotiated',False) else None
				if consolidate:
					self.peer.reactor.processes.message(msg_id,self.neighbor,'receive',message,negotiated,header,body)
				elif parsed:
					self.peer.reactor.processes.message(msg_id,self.neighbor,'receive',message,negotiated,b'',b'')

			if message.TYPE == Notification.TYPE:
				raise message

			if message.TYPE == Update.TYPE and Attribute.CODE.INTERNAL_DISCARD in message.attributes:
				yield _NOP
			else:
				yield message

	def validate_open (self):
		error = self.negotiated.validate(self.neighbor)
		if error is not None:
			raise Notify(*error)

		if self.neighbor.api['negotiated']:
			self.peer.reactor.processes.negotiated(self.peer.neighbor,self.negotiated)

		if self.negotiated.mismatch:
			self.logger.warning('--------------------------------------------------------------------',self.connection.session())
			self.logger.warning('the connection can not carry the following family/families',self.connection.session())
			for reason,(afi,safi) in self.negotiated.mismatch:
				self.logger.warning(' - %s is not configured for %s/%s' % (reason,afi,safi),self.connection.session())
			self.logger.warning('therefore no routes of this kind can be announced on the connection',self.connection.session())
			self.logger.warning('--------------------------------------------------------------------',self.connection.session())

	def read_open (self, ip):
		for received_open in self.read_message():
			if received_open.TYPE == NOP.TYPE:
				yield received_open
			else:
				break

		if received_open.TYPE != Open.TYPE:
			raise Notify(5,1,'The first packet received is not an open message (%s)' % received_open)

		self.logger.debug('<< %s' % received_open,self.connection.session())
		yield received_open

	def read_keepalive (self):
		for message in self.read_message():
			if message.TYPE == NOP.TYPE:
				yield message
			else:
				break

		if message.TYPE != KeepAlive.TYPE:
			raise Notify(5,2)

		yield message

	#
	# Sending message to peer
	#

	def new_open (self):
		if self.neighbor.local_as:
			local_as = self.neighbor.local_as
		elif self.negotiated.received_open:
			local_as = self.negotiated.received_open.asn
		else:
			raise RuntimeError('no ASN available for the OPEN message')

		sent_open = Open(
			Version(4),
			local_as,
			self.neighbor.hold_time,
			self.neighbor.router_id,
			Capabilities().new(self.neighbor,self.peer._restarted)
		)

		# we do not buffer open message in purpose
		for _ in self.write(sent_open):
			yield _NOP

		self.logger.debug('>> %s' % sent_open,self.connection.session())
		yield sent_open

	def new_keepalive (self, comment=''):
		keepalive = KeepAlive()

		for _ in self.write(keepalive):
			yield _NOP

		self.logger.debug('>> KEEPALIVE%s' % (' (%s)' % comment if comment else ''),self.connection.session())

		yield keepalive

	def new_notification (self, notification):
		for _ in self.write(notification):
			yield _NOP
		self.logger.debug('>> NOTIFICATION (%d,%d,"%s")' % (notification.code,notification.subcode,notification.data),self.connection.session())
		yield notification

	def new_update (self, include_withdraw):
		updates = self.neighbor.rib.outgoing.updates(self.neighbor.group_updates)
		number = 0
		for update in updates:
			for message in update.messages(self.negotiated,include_withdraw):
				number += 1
				for boolean in self.send(message):
					# boolean is a transient network error we already announced
					yield _NOP
		if number:
			self.logger.debug('>> %d UPDATE(s)' % number,self.connection.session())
		yield _UPDATE

	def new_eor (self, afi, safi):
		eor = EOR(afi,safi)
		for _ in self.write(eor):
			yield _NOP
		self.logger.debug('>> EOR %s %s' % (afi,safi),self.connection.session())
		yield eor

	def new_eors (self, afi=AFI.undefined,safi=SAFI.undefined):
		# Send EOR to let our peer know he can perform a RIB update
		if self.negotiated.families:
			families = self.negotiated.families if (afi,safi) == (AFI.undefined,SAFI.undefined) else [(afi,safi),]
			for eor_afi,eor_safi in families:
				for _ in self.new_eor(eor_afi,eor_safi):
					yield _
		else:
			# If we are not sending an EOR, send a keepalive as soon as when finished
			# So the other routers knows that we have no (more) routes to send ...
			# (is that behaviour documented somewhere ??)
			for eor in self.new_keepalive('EOR'):
				yield _NOP
			yield _UPDATE

	def new_operational (self, operational, negotiated):
		for _ in self.write(operational,negotiated):
			yield _NOP
		self.logger.debug('>> OPERATIONAL %s' % str(operational),self.connection.session())
		yield operational

	def new_refresh (self, refresh):
		for _ in self.write(refresh,None):
			yield _NOP
		self.logger.debug('>> REFRESH %s' % str(refresh),self.connection.session())
		yield refresh



# http://teethgrinder.co.uk/perm.php?a=Python-memory-leak-detector
import gc
import inspect


def dump():
	# force collection
	print("\nCollecting GARBAGE:")
	gc.collect()
	# prove they have been collected
	print("\nCollecting GARBAGE:")
	gc.collect()

	print("\nGARBAGE OBJECTS:")
	for x in gc.garbage:
		s = str(x)
		if len(s) > 80:
			s = "%s..." % s[:80]

		print("::", s)
		print("		type:", type(x))
		print("   referrers:", len(gc.get_referrers(x)))
		try:
			print("	is class:", inspect.isclass(type(x)))
			print("	  module:", inspect.getmodule(x))

			lines, line_num = inspect.getsourcelines(type(x))
			print("	line num:", line_num)
			for l in lines:
				print("		line:", l.rstrip("\n"))
		except Exception:
			pass

		print()



#!/usr/bin/env python
# encoding: utf-8
"""
response/text.py

Created by Thomas Mangin on 2012-12-30.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import os

from exabgp.util import hexstring
from exabgp.bgp.message import IN


class Text (object):
	def __init__ (self, version):
		self.version = version

	def _header_body (self, header, body):
		header = ' header %s' % hexstring(header) if header else ''
		body = ' body %s' % hexstring(body) if body else ''

		total_string = header+body if body else header

		return total_string

	def _counter (self, neighbor):
		return None

	def up (self, neighbor):
		return 'neighbor %s up\n' % (
			neighbor.peer_address
		)

	def connected (self, neighbor):
		return 'neighbor %s connected\n' % (
			neighbor.peer_address
		)

	def down (self, neighbor, reason=''):
		return 'neighbor %s down - %s\n' % (
			neighbor.peer_address,
			reason
		)

	def shutdown (self):
		return 'shutdown %d %d\n' % (
			os.getpid(),
			os.getppid()
		)

	def negotiated (self, neighbor, negotiated):
		return None

	def fsm (self, neighbor, fsm):
		return None

	def signal (self, neighbor, signal):
		return None

	def notification (self, neighbor, direction, message, negotiated, header, body):
		return 'neighbor %s %s notification code %d subcode %d data %s%s\n' % (
			neighbor.peer_address,
			direction,
			message.code,
			message.subcode,
			hexstring(message.data),
			self._header_body(header,body)
		)

	def packets (self, neighbor, direction, category, negotiated, header, body):
		return 'neighbor %s %s %d%s\n' % (
			neighbor.peer_address,
			direction,
			category,
			self._header_body(header,body)
		)

	def keepalive (self, neighbor, direction, negotiated, header, body):
		return 'neighbor %s %s keepalive%s\n' % (
			neighbor.peer_address,
			direction,
			self._header_body(header,body)
		)

	def open (self, neighbor, direction, sent_open, negotiated, header, body):
		return 'neighbor %s %s open version %d asn %d hold_time %s router_id %s capabilities [%s]%s\n' % (
			neighbor.peer_address,
			direction,
			sent_open.version,
			sent_open.asn,
			sent_open.hold_time,
			sent_open.router_id,
			str(sent_open.capabilities).lower(),
			self._header_body(header,body)
		)

	def update (self, neighbor, direction, update, negotiated, header, body):
		prefix = 'neighbor %s %s update' % (
			neighbor.peer_address,
			direction,
		)

		r = '%s start\n' % prefix

		attributes = str(update.attributes)
		for nlri in update.nlris:
			if nlri.EOR:
				r += '%s route %s\n' % (prefix,nlri.extensive())
			elif nlri.action == IN.ANNOUNCED:  # pylint: disable=E1101
				if nlri.nexthop:
					r += '%s announced %s%s\n' % (prefix,nlri.extensive(),attributes)
				else:
					# This is an EOR or Flow or ... something newer
					r += '%s %s %s\n' % (prefix,nlri.extensive(),attributes)
			else:
				r += '%s withdrawn %s\n' % (prefix,nlri.extensive())
		if header or body:
			r += '%s\n' % self._header_body(header,body)

		r += '%s end\n' % prefix

		return r

	def refresh (self, neighbor, direction, refresh, negotiated, header, body):
		return 'neighbor %s %s route-refresh afi %s safi %s %s%s\n' % (
			neighbor.peer_address,
			direction,
			refresh.afi,
			refresh.safi,
			refresh.reserved,
			self._header_body(header,body)
		)

	def _operational_advisory (self, neighbor, direction, operational, header, body):
		return 'neighbor %s %s operational %s afi %s safi %s advisory "%s"%s' % (
			neighbor.peer_address,
			direction,
			operational.name,
			operational.afi,
			operational.safi,
			operational.data,
			self._header_body(header,body)
		)

	def _operational_query (self, neighbor, direction, operational, header, body):
		return 'neighbor %s %s operational %s afi %s safi %s%s' % (
			neighbor.peer_address,
			direction,
			operational.name,
			operational.afi,
			operational.safi,
			self._header_body(header,body)
		)

	def _operational_counter (self, neighbor, direction, operational, header, body):
		return 'neighbor %s %s operational %s afi %s safi %s router-id %s sequence %d counter %d%s' % (
			neighbor.peer_address,
			direction,
			operational.name,
			operational.afi,
			operational.safi,
			operational.routerid,
			operational.sequence,
			operational.counter,
			self._header_body(header,body)
		)

	def operational (self, neighbor, direction, what, operational, negotiated, header, body):
		if what == 'advisory':
			return self._operational_advisory(neighbor,direction,operational,header,body)
		elif what == 'query':
			return self._operational_query(neighbor,direction,operational,header,body)
		elif what == 'counter':
			return self._operational_counter(neighbor,direction,operational,header,body)
		# elif what == 'interface':
		# 	return self._operational_interface(peer,operational)
		else:
			raise RuntimeError('the code is broken, we are trying to print a unknown type of operational message')



"""
process.py

Created by Thomas Mangin on 2011-05-02.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import os
import errno
import time
import subprocess
import select
import fcntl

from exabgp.util import str_ascii
from exabgp.util import bytes_ascii
from exabgp.util.errstr import errstr
from exabgp.reactor.network.error import error

from exabgp.configuration.core.format import formated
from exabgp.reactor.api.response import Response
from exabgp.reactor.api.response.answer import Answer

from exabgp.bgp.message import Message
from exabgp.logger import Logger

from exabgp.version import json as json_version
from exabgp.version import text as text_version

from exabgp.configuration.environment import environment


# pylint: disable=no-self-argument,not-callable,unused-argument,invalid-name

class ProcessError (Exception):
	pass


def preexec_helper ():
	# make this process a new process group
	# os.setsid()
	# This prevent the signal to be sent to the children (and create a new process group)
	os.setpgrp()
	# signal.signal(signal.SIGINT, signal.SIG_IGN)


class Processes (object):
	# how many time can a process can respawn in the time interval
	respawn_timemask = 0xFFFFFF - 0b111111
	# '0b111111111111111111000000' (around a minute, 63 seconds)

	_dispatch = {}

	def __init__ (self):
		self.logger = Logger()
		self.clean()
		self.silence = False
		self._buffer = {}
		self._configuration = {}

		self.respawn_number = 5 if environment.settings().api.respawn else 0
		self.terminate_on_error = environment.settings().api.terminate
		self.ack = environment.settings().api.ack

	def number (self):
		return len(self._process)

	def clean (self):
		self._process = {}
		self._encoder = {}
		self._broken = []
		self._respawning = {}

	def _handle_problem (self, process):
		if self.respawn_number:
			self.logger.debug('issue with the process, restarting it','process')
			self._terminate(process)
			self._start(process)
		else:
			self.logger.debug('issue with the process, terminating it','process')
			self._terminate(process)

	def _terminate (self, process):
		self.logger.debug('terminating process %s' % process,'process')
		try:
			self._process[process].terminate()
		except OSError:
			# the process is most likely already dead
			pass
		self._process[process].wait()
		del self._process[process]

	def terminate (self):
		for process in list(self._process):
			if not self.silence:
				try:
					self._write(process,self._encoder[process].shutdown())
				except ProcessError:
					pass
		self.silence = True
		# waiting a little to make sure IO is flushed to the pipes
		# we are using unbuffered IO but still ..
		time.sleep(0.1)
		for process in list(self._process):
			try:
				self._terminate(process)
			except OSError:
				# we most likely received a SIGTERM signal and our child is already dead
				self.logger.debug('child process %s was already dead' % process,'process')
		self.clean()

	def _start (self,process):
		try:
			if process in self._process:
				self.logger.debug('process already running','process')
				return
			if process not in self._configuration:
				self.logger.debug('can not start process, no configuration for it','process')
				return
			# Prevent some weird termcap data to be created at the start of the PIPE
			# \x1b[?1034h (no-eol) (esc)
			os.environ['TERM'] = 'dumb'

			configuration = self._configuration[process]

			run = configuration.get('run','')
			if run:
				api = configuration.get('encoder','')
				self._encoder[process] = Response.Text(text_version) if api == 'text' else Response.JSON(json_version)

				self._process[process] = subprocess.Popen(
					run,
					stdin=subprocess.PIPE,
					stdout=subprocess.PIPE,
					preexec_fn=preexec_helper
					# This flags exists for python 2.7.3 in the documentation but on on my MAC
					# creationflags=subprocess.CREATE_NEW_PROCESS_GROUP
				)
				fcntl.fcntl(self._process[process].stdout.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)

				self.logger.debug('forked process %s' % process,'process')

				around_now = int(time.time()) & self.respawn_timemask
				if process in self._respawning:
					if around_now in self._respawning[process]:
						self._respawning[process][around_now] += 1
						# we are respawning too fast
						if self._respawning[process][around_now] > self.respawn_number:
							self.logger.critical(
								'Too many death for %s (%d) terminating program' % (process,self.respawn_number),
								'process'
							)
							raise ProcessError()
					else:
						# reset long time since last respawn
						self._respawning[process] = {around_now: 1}
				else:
					# record respawing
					self._respawning[process] = {around_now: 1}

		except (subprocess.CalledProcessError,OSError,ValueError) as exc:
			self._broken.append(process)
			self.logger.debug('could not start process %s' % process,'process')
			self.logger.debug('reason: %s' % str(exc),'process')

	def start (self, configuration, restart=False):
		for process in list(self._process):
			if process not in configuration:
				self._terminate(process)
		self._configuration = configuration
		for process in configuration:
			if restart and process in list(self._process):
				self._terminate(process)
			self._start(process)

	def broken (self, neighbor):
		if self._broken:
			for process in self._configuration:
				if process in self._broken:
					return True
		return False

	def fds (self):
		return [self._process[process].stdout for process in self._process]

	def received (self):
		consumed_data = False

		for process in list(self._process):
			try:
				proc = self._process[process]
				poll = proc.poll()
				# proc.poll returns None if the process is still fine
				# -[signal], like -15, if the process was terminated
				if poll is not None:
					self._handle_problem(process)
					return
				r,_,_ = select.select([proc.stdout,],[],[],0)
				if not r:
					continue
				try:
					# Calling next() on Linux and OSX works perfectly well
					# but not on OpenBSD where it always raise StopIteration
					# and only readline() works
					buf = str_ascii(proc.stdout.read(16384))
					if buf == '' and poll is not None:
						# if proc.poll() is None then
						# process is fine, we received an empty line because
						# we're doing .readline() on a non-blocking pipe and
						# the process maybe has nothing to send yet
						self._handle_problem(process)
						continue

					raw = self._buffer.get(process,'') + buf

					while '\n' in raw:
						line,raw = raw.split('\n',1)
						line = line.rstrip()
						consumed_data = True
						self.logger.debug('command from process %s : %s ' % (process,line),'process')
						yield (process,formated(line))

					self._buffer[process] = raw

				except IOError as exc:
					if not exc.errno or exc.errno in error.fatal:
						# if the program exits we can get an IOError with errno code zero !
						self._handle_problem(process)
					elif exc.errno in error.block:
						# we often see errno.EINTR: call interrupted and
						# we most likely have data, we will try to read them a the next loop iteration
						pass
					else:
						self.logger.debug('unexpected errno received from forked process (%s)' % errstr(exc),'process')
				except StopIteration:
					if not consumed_data:
						self._handle_problem(process)
			except (subprocess.CalledProcessError,OSError,ValueError):
				self._handle_problem(process)

	def _write (self, process, string, neighbor=None):
		if string is None:
			return True

		# XXX: FIXME: This is potentially blocking
		while True:
			try:
				self._process[process].stdin.write(bytes_ascii('%s\n' % string))
			except IOError as exc:
				self._broken.append(process)
				if exc.errno == errno.EPIPE:
					self._broken.append(process)
					self.logger.debug('issue while sending data to our helper program','process')
					raise ProcessError()
				else:
					# Could it have been caused by a signal ? What to do.
					self.logger.debug('error received while sending data to helper program, retrying (%s)' % errstr(exc),'process')
					continue
			break

		try:
			self._process[process].stdin.flush()
		except IOError as exc:
			# AFAIK, the buffer should be flushed at the next attempt.
			self.logger.debug('error received while FLUSHING data to helper program, retrying (%s)' % errstr(exc),'process')

		return True

	def answer (self, service, string, force=False):
		if force or self.ack:
			self.logger.debug('responding to %s : %s' % (service,string.replace('\n', '\\n')), 'process')
			self._write(service,string)

	def answer_done (self, service):
		self.answer(service,Answer.done)

	def answer_error (self, service):
		self.answer(service,Answer.error)

	def _notify (self, neighbor, event):
		for process in neighbor.api[event]:
			yield process

	# do not do anything if silenced
	# no-self-argument

	def silenced (function):
		def closure (self, *args):
			if self.silence:
				return
			return function(self,*args)
		return closure

	# invalid-name
	@silenced
	def up (self, neighbor):
		for process in self._notify(neighbor,'neighbor-changes'):
			self._write(process,self._encoder[process].up(neighbor),neighbor)

	@silenced
	def connected (self, neighbor):
		for process in self._notify(neighbor,'neighbor-changes'):
			self._write(process,self._encoder[process].connected(neighbor),neighbor)

	@silenced
	def down (self, neighbor, reason):
		for process in self._notify(neighbor,'neighbor-changes'):
			self._write(process,self._encoder[process].down(neighbor,reason),neighbor)

	@silenced
	def negotiated (self, neighbor, negotiated):
		for process in self._notify(neighbor,'negotiated'):
			self._write(process,self._encoder[process].negotiated(neighbor,negotiated),neighbor)

	@silenced
	def fsm (self, neighbor, fsm):
		for process in self._notify(neighbor,'fsm'):
			self._write(process,self._encoder[process].fsm(neighbor,fsm),neighbor)

	@silenced
	def signal (self, neighbor, signal):
		for process in self._notify(neighbor,'signal'):
			self._write(process,self._encoder[process].signal(neighbor,signal),neighbor)

	@silenced
	def packets (self, neighbor, direction, category, header, body):
		for process in self._notify(neighbor,'%s-packets' % direction):
			self._write(process,self._encoder[process].packets(neighbor,direction,category,header,body),neighbor)

	@silenced
	def notification (self, neighbor, direction, code, subcode, data, header, body):
		for process in self._notify(neighbor,'neighbor-changes'):
			self._write(process,self._encoder[process].notification(neighbor,direction,code,subcode,data,header,body),neighbor)

	@silenced
	def message (self, message_id, neighbor, direction, message, negotiated, header, *body):
		self._dispatch[message_id](self,neighbor,direction,message,negotiated,header,*body)

	# registering message functions
	# no-self-argument

	def register_process (message_id, storage=_dispatch):
		def closure (function):
			def wrap (*args):
				function(*args)
			storage[message_id] = wrap
			return wrap
		return closure

	# notifications are handled in the loop as they use different arguments

	@register_process(Message.CODE.OPEN)
	def _open (self, peer, direction, message, negotiated, header, body):
		for process in self._notify(peer,'%s-%s' % (direction,Message.CODE.OPEN.SHORT)):
			self._write(process,self._encoder[process].open(peer,direction,message,negotiated,header,body),peer)

	@register_process(Message.CODE.UPDATE)
	def _update (self, peer, direction, update, negotiated, header, body):
		for process in self._notify(peer,'%s-%s' % (direction,Message.CODE.UPDATE.SHORT)):
			self._write(process,self._encoder[process].update(peer,direction,update,negotiated,header,body),peer)

	@register_process(Message.CODE.NOTIFICATION)
	def _notification (self, peer, direction, message, negotiated, header, body):
		for process in self._notify(peer,'%s-%s' % (direction,Message.CODE.NOTIFICATION.SHORT)):
			self._write(process,self._encoder[process].notification(peer,direction,message,negotiated,header,body),peer)

	# unused-argument, must keep the API
	@register_process(Message.CODE.KEEPALIVE)
	def _keepalive (self, peer, direction, keepalive, negotiated, header, body):
		for process in self._notify(peer,'%s-%s' % (direction,Message.CODE.KEEPALIVE.SHORT)):
			self._write(process,self._encoder[process].keepalive(peer,direction,negotiated,header,body),peer)

	@register_process(Message.CODE.ROUTE_REFRESH)
	def _refresh (self, peer, direction, refresh, negotiated, header, body):
		for process in self._notify(peer,'%s-%s' % (direction,Message.CODE.ROUTE_REFRESH.SHORT)):
			self._write(process,self._encoder[process].refresh(peer,direction,refresh,negotiated,header,body),peer)

	@register_process(Message.CODE.OPERATIONAL)
	def _operational (self, peer, direction, operational, negotiated, header, body):
		for process in self._notify(peer,'%s-%s' % (direction,Message.CODE.OPERATIONAL.SHORT)):
			self._write(process,self._encoder[process].operational(peer,direction,operational.category,operational,negotiated,header,body),peer)



# encoding: utf-8
"""
neighbor/parser.py

Created by Thomas Mangin on 2014-07-01.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import socket
from string import ascii_letters
from string import digits

from exabgp.bgp.message.open.routerid import RouterID
from exabgp.bgp.message.open.holdtime import HoldTime
from exabgp.configuration.parser import string
from exabgp.protocol.ip import IP


def inherit (tokeniser):
	if len(tokeniser.tokens) == 2:
		return [tokeniser()]
	if len(tokeniser.tokens) < 4 or tokeniser.tokens[1] != '[' or tokeniser.tokens[-1] != ']':
		raise ValueError('invalid inherit list')
	return tokeniser.tokens[2:-1]


def hostname (tokeniser):
	value = string(tokeniser)
	if not value[0].isalnum():
		raise ValueError('bad host-name (alphanumeric)')
	if not value[-1].isalnum() or value[-1].isdigit():
		raise ValueError('bad host-name (alphanumeric)')
	if '..' in value:
		raise ValueError('bad host-name (double period)')
	if not all(True if c in ascii_letters + digits + '.-' else False for c in value):
		raise ValueError('bad host-name (charset)')
	if len(value) > 255:
		raise ValueError('bad host-name (length)')

	return value


def domainname (tokeniser):
	value = string(tokeniser)
	if not value:
		raise ValueError('bad domain-name')
	if not value[0].isalnum() or value[0].isdigit():
		raise ValueError('bad domain-name')
	if not value[-1].isalnum() or value[-1].isdigit():
		raise ValueError('bad domain-name')
	if '..' in value:
		raise ValueError('bad domain-name')
	if not all(True if c in ascii_letters + digits + '.-' else False for c in value):
		raise ValueError('bad domain-name')
	if len(value) > 255:
		raise ValueError('bad domain-name (length)')
	return value


def description (tokeniser):
	try:
		return string(tokeniser)
	except Exception:
		raise ValueError('bad neighbor description')


def md5 (tokeniser):
	value = tokeniser()
	if not value:
		raise ValueError('value requires the value password as an argument (quoted or unquoted).  FreeBSD users should use "kernel" as the argument.')
	return value


def ttl (tokeniser):
	value = tokeniser()
	try:
		attl = int(value)
	except ValueError:
		if value in ('false','disable','disabled'):
			return None
		raise ValueError('invalid ttl-security "%s"' % value)
	if attl < 0:
		raise ValueError('ttl-security can not be negative')
	if attl > 255:
		raise ValueError('ttl must be smaller than 256')
	return attl


def local_address(tokeniser):
	if not tokeniser.tokens:
		raise ValueError("an ip address  or 'auto' is required")

	value = tokeniser()
	if value == 'auto':
		return None
	try:
		return IP.create(value)
	except (IndexError,ValueError,socket.error):
		raise ValueError('"%s" is an invalid IP address' % value)


def router_id (tokeniser):
	value = tokeniser()
	try:
		return RouterID(value)
	except ValueError:
		raise ValueError ('"%s" is an invalid router-id' % value)


def hold_time (tokeniser):
	value = tokeniser()
	try:
		holdtime = HoldTime(int(value))
	except ValueError:
		raise ValueError ('"%s" is an invalid hold-time' % value)
	if holdtime < 3 and holdtime != 0:
		raise ValueError('holdtime must be zero or at least three seconds')
	if holdtime > HoldTime.MAX:
		raise ValueError('holdtime must be smaller or equal to %d' % HoldTime.MAX)
	return holdtime


def processes (tokeniser):
	result = []
	token = tokeniser()
	if token != '[':
		raise ValueError('invalid processes, does not start with [')

	while True:
		token = tokeniser()
		if not token:
			raise ValueError('invalid processes, does not end with ]')
		if token == ']':
			break
		if token == ',':
			continue
		result.append(token)

	return result



# encoding: utf-8
"""
parse_process.py

Created by Thomas Mangin on 2015-06-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import time
import copy
from collections import defaultdict

from exabgp.configuration.core import Section
from exabgp.configuration.parser import boolean
from exabgp.configuration.neighbor.parser import processes


class _ParseDirection (Section):
	action = {
		'parsed':       'set-command',
		'packets':      'set-command',
		'consolidate':  'set-command',
		'open':         'set-command',
		'update':       'set-command',
		'notification': 'set-command',
		'keepalive':    'set-command',
		'refresh':      'set-command',
		'operational':  'set-command',
	}

	known = {
		'parsed':       boolean,
		'packets':      boolean,
		'consolidate':  boolean,
		'open':         boolean,
		'update':       boolean,
		'notification': boolean,
		'keepalive':    boolean,
		'refresh':      boolean,
		'operational':  boolean,
	}

	default = {
		'parsed':       True,
		'packets':      True,
		'consolidate':  True,
		'open':         True,
		'update':       True,
		'notification': True,
		'keepalive':    True,
		'refresh':      True,
		'operational':  True,
	}

	syntax = '{\n  %s;\n}' % ';\n  '.join(default.keys())

	def __init__ (self, tokeniser, scope, error, logger):
		Section.__init__(self,tokeniser,scope,error,logger)

	def clear (self):
		pass

	def pre (self):
		return True

	def post (self):
		return True


class ParseSend (_ParseDirection):
	syntax = \
		'send %s' % _ParseDirection.syntax

	name = 'api/send'


class ParseReceive (_ParseDirection):
	syntax = \
		'receive %s' % _ParseDirection.syntax

	name = 'api/receive'


class ParseAPI (Section):
	syntax = \
		'process {\n' \
		'  processes [ name-of-processes ];\n' \
		'  neighbor-changes;\n' \
		'  %s\n' \
		'  %s\n' \
		'}' % (
			'\n  '.join(ParseSend.syntax.split('\n')),
			'\n  '.join(ParseReceive.syntax.split('\n'))
		)

	known = {
		'processes':        processes,
		'neighbor-changes': boolean,
		'negotiated':       boolean,
		'fsm':              boolean,
		'signal':           boolean,
	}

	action = {
		'processes':        'set-command',
		'neighbor-changes': 'set-command',
		'negotiated':       'set-command',
		'fsm':              'set-command',
		'signal':           'set-command',
	}

	default = {
		'neighbor-changes': True,
		'negotiated':       True,
		'fsm':              True,
		'signal':           True,
	}

	DEFAULT_API = {
		'neighbor-changes': [],
		'negotiated':       [],
		'fsm':              [],
		'signal':           [],
		'processes':        [],
	}

	name = 'api'

	def __init__ (self, tokeniser, scope, error, logger):
		Section.__init__(self,tokeniser,scope,error,logger)
		self.api = {}
		self.named = ''

	@classmethod
	def _empty (cls):
		return copy.deepcopy(cls.DEFAULT_API)

	def clear (self):
		self.api = {}
		self.named = ''
		pass

	def pre (self):
		named = self.tokeniser.iterate()
		self.named = named if named else 'auto-named-%d' % int(time.time()*1000000)
		self.check_name(self.named)
		self.scope.enter(self.named)
		self.scope.to_context()
		return True

	def post (self):
		self.scope.leave()
		self.scope.to_context()
		return True

	@classmethod
	def flatten (cls,apis):
		built = cls._empty()

		for api in apis.values():
			procs = api.get('processes',[])

			built.setdefault('processes',[]).extend(procs)

			for command in ('neighbor-changes','negotiated','fsm','signal'):
				built.setdefault(command,[]).extend(procs if api.get(command,False) else [])

			for direction in ('send','receive'):
				data = api.get(direction,{})
				for action in ('parsed','packets','consolidate','open', 'update', 'notification', 'keepalive', 'refresh', 'operational'):
					built.setdefault("%s-%s" % (direction,action),[]).extend(procs if data.get(action,False) else [])

		return built


for way in ('send','receive'):
	for name in ('parsed','packets','consolidate','open', 'update', 'notification', 'keepalive', 'refresh', 'operational'):
		ParseAPI.DEFAULT_API["%s-%s" % (way,name)] = []



# encoding: utf-8
"""
template.py

Created by Thomas Mangin on 2015-06-16.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.configuration.core import Section
from exabgp.configuration.neighbor import ParseNeighbor

from exabgp.configuration.neighbor.api import ParseAPI


class ParseTemplate (Section):
	syntax = ''

	name = 'template'

	def __init__ (self, tokeniser, scope, error, logger):
		Section.__init__(self,tokeniser,scope,error,logger)

	def clear (self):
		self._names = []

	def pre (self):
		return True

	def post (self):
		return True
